@@ -13,8 +13,8 @@
 #include <sys/poll.h>
 #include <sys/ipc.h>
 #include <sys/sem.h>
-#include "sg_include.h"
 #include "sg_lib.h"
+#include "sg_io_linux.h"
 
 /* Test code for D. Gilbert's extensions to the Linux OS SCSI generic ("sg")
    device driver.
