@@ -17,8 +17,8 @@
 #include <linux/major.h>
 #include <sys/time.h>
 typedef unsigned char u_char;   /* horrible, for scsi.h */
-#include "sg_include.h"
 #include "sg_lib.h"
+#include "sg_io_linux.h"
 #include "llseek.h"
 
 /* A utility program for the Linux OS SCSI generic ("sg") device driver.
@@ -301,7 +301,7 @@ int read_capacity(int sg_fd, int * num_sect, int * sect_sz)
         return -1;
     }
     res = sg_err_category3(&io_hdr);
-    if (SG_LIB_CAT_MEDIA_CHANGED == res)
+    if (SG_LIB_CAT_UNIT_ATTENTION == res)
         return 2; /* probably have another go ... */
     else if (SG_LIB_CAT_CLEAN != res) {
         sg_chk_n_print3("read capacity", &io_hdr, 1);
@@ -528,7 +528,7 @@ int sg_finish_io(int wr, Rq_elem * rep)
             fprintf(stderr, "Recovered error on block=%d, num=%d\n",
                     rep->blk, rep->num_blks);
             break;
-        case SG_LIB_CAT_MEDIA_CHANGED:
+        case SG_LIB_CAT_UNIT_ATTENTION:
             return 1;
         default:
             {
