@@ -9,8 +9,8 @@
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include "sg_include.h"
 #include "sg_lib.h"
+#include "sg_linux_inc.h"
 
 /* A utility program for the Linux OS SCSI generic ("sg") device driver.
 *  Copyright (C) 1999 D. Gilbert and P. Allworth
@@ -143,7 +143,7 @@ int read_capacity(int sg_fd, int * num_sect, int * sect_sz)
         return -1;
     }
     res = sg_err_category3(&io_hdr);
-    if (SG_LIB_CAT_MEDIA_CHANGED == res)
+    if (SG_LIB_CAT_UNIT_ATTENTION == res)
         return 2; /* probably have another go ... */
     else if (SG_LIB_CAT_CLEAN != res) {
         sg_chk_n_print3("read capacity", &io_hdr);
@@ -259,7 +259,7 @@ int sg_finish_io(Rq_coll * clp, int wr, Rq_elem ** repp)
             printf("Recovered error on block=%d, num=%d\n",
                    rep->blk, rep->num_blks);
             break;
-        case SG_LIB_CAT_MEDIA_CHANGED:
+        case SG_LIB_CAT_UNIT_ATTENTION:
             return 1;
         default:
             sg_chk_n_print3(rep->wr ? "writing": "reading", hp);
