@@ -7,8 +7,8 @@
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include "sg_include.h"
 #include "sg_lib.h"
+#include "sg_io_linux.h"
 
 /* This program performs a READ_16 command as scsi mid-level support
    16 byte commands from lk 2.4.15
