@@ -6,7 +6,7 @@
 **
 ** Copyright (C) 2003  Grant Grundler    grundler at parisc-linux dot org
 ** Copyright (C) 2003  James Bottomley       jejb at parisc-linux dot org
-** Copyright (C) 2005  Douglas Gilbert   dgilbert at interlog dot com
+** Copyright (C) 2005-2006  Douglas Gilbert   dgilbert at interlog dot com
 **
 **   This program is free software; you can redistribute it and/or modify
 **   it under the terms of the GNU General Public License as published by
@@ -35,40 +35,25 @@
 #include <string.h>
 #include <getopt.h>
 #include <unistd.h>
-#include <sys/fcntl.h>
-#include <sys/ioctl.h>
-#include <scsi/scsi.h>
-#include <scsi/scsi_ioctl.h>
-#include <scsi/sg.h>
-#include <sys/errno.h>
 
 #include "sg_lib.h"
 #include "sg_cmds.h"
 
+
 #define RW_ERROR_RECOVERY_PAGE 1  /* every disk should have one */
 #define FORMAT_DEV_PAGE 3         /* Format Device Mode Page [now obsolete] */
 #define CONTROL_MODE_PAGE 0xa     /* alternative page all devices have?? */
 
-#define CDB_SIZE                6       /* SCSI Command Block */
-#define MODE_HDR_SIZE           4       /* Mode Sense Header */
-#define BLOCK_DESCR_SIZE        8       /* Block Descriptor Header */
-
-#define LOGICAL_UNIT_NOT_READY  4 /* ASC */
-#define FORMAT_IN_PROGRESS      4 /* ASCQ */
-
-#define SHORT_TIMEOUT           20000   /* 20 seconds unless immed=0 ... */
-#define FORMAT_TIMEOUT          (4 * 3600 * 1000)       /* 4 hours ! */
+#define SHORT_TIMEOUT           20   /* 20 seconds unless immed=0 ... */
+#define FORMAT_TIMEOUT          (4 * 3600)       /* 4 hours ! */
 
 #define POLL_DURATION_SECS 30
  
 
-#define MAX_SENSE_SZ    32
-static unsigned char sbuff[MAX_SENSE_SZ];
-
 #define MAX_BUFF_SZ     252
 static unsigned char dbuff[MAX_BUFF_SZ];
 
-static char * version_str = "1.05 20050908";
+static char * version_str = "1.09 20060623";
 
 static struct option long_options[] = {
         {"count", 1, 0, 'c'},
@@ -79,6 +64,7 @@ static struct option long_options[] = {
         {"pinfo", 0, 0, 'p'},
         {"resize", 0, 0, 'r'},
         {"rto_req", 0, 0, 'R'},
+        {"six", 0, 0, '6'},
         {"size", 1, 0, 's'},
         {"verbose", 0, 0, 'v'},
         {"version", 0, 0, 'V'},
@@ -86,180 +72,71 @@ static struct option long_options[] = {
         {0, 0, 0, 0},
 };
 
-static const char * scsi_ptype_strs[] = {
-        "disk",                             /* 0x0 */
-        "tape",
-        "printer",
-        "processor",
-        "write once optical disk",
-        "cd/dvd",
-        "scanner",
-        "optical memory device",
-        "medium changer",                   /* 0x8 */
-        "communications",
-        "graphics [0xa]",
-        "graphics [0xb]",
-        "storage array controller",
-        "enclosure services device",
-        "simplified direct access device",
-        "optical card reader/writer device",
-        "bridge controller commands",       /* 0x10 */
-        "object storage device",
-        "automation/drive interface",
-        "0x13", "0x14", "0x15", "0x16", "0x17", "0x18",
-        "0x19", "0x1a", "0x1b", "0x1c", "0x1d",
-        "well known logical unit",
-        "no physical device on this lu",
-};
-
-static const char * get_ptype_str(int scsi_ptype)
-{
-        int num = sizeof(scsi_ptype_strs) / sizeof(scsi_ptype_strs[0]);
-
-        return (scsi_ptype < num) ? scsi_ptype_strs[scsi_ptype] : "";
-}
-
-
-/* Return 0 on success, else -1 */
+/* Return 0 on success, else see sg_ll_format_unit() */
 static int
 scsi_format(int fd, int pinfo, int rto_req, int immed, int early, int verbose)
 {
-        int k, res;
+        int res;
         const char FORMAT_HEADER_SIZE = 4;
-        unsigned char cdb[CDB_SIZE], fmt_hdr[FORMAT_HEADER_SIZE];
-        sg_io_hdr_t io_hdr;
-
-        cdb[0] = FORMAT_UNIT;
-        cdb[1] = (pinfo ? 0x80 : 0) | (rto_req ? 0x40 : 0) |
-                 (immed ? 0x10 : 0);
-        cdb[2] = 0;             /* vendor specific */
-        cdb[3] = 0;             /* interleave MSB */
-        cdb[4] = 0;             /* interleave LSB */
-        cdb[5] = 0;             /* control */
+        unsigned char fmt_hdr[FORMAT_HEADER_SIZE];
 
         /* fmt_hdr is a short format header, only used when 'immed' is set */
         fmt_hdr[0] = 0;         /* reserved */
-        fmt_hdr[1] = 0x02;      /* use device defaults, IMMED return */
+        fmt_hdr[1] = 0x2;       /* use device defaults, IMMED return */
         fmt_hdr[2] = 0;         /* defect list length MSB */
         fmt_hdr[3] = 0;         /* defect list length LSB */
 
-        memset(&io_hdr, 0, sizeof(sg_io_hdr_t));
-        memset(sbuff, 0, MAX_SENSE_SZ);
-        io_hdr.interface_id = 'S';
-        io_hdr.dxfer_direction = immed ? SG_DXFER_TO_DEV : SG_DXFER_NONE;
-        io_hdr.cmd_len = CDB_SIZE;
-        io_hdr.mx_sb_len = MAX_SENSE_SZ;
-        io_hdr.iovec_count = 0;         /* no scatter gather */
-        if (immed) {
-                io_hdr.dxfer_len = FORMAT_HEADER_SIZE;
-                io_hdr.dxferp = fmt_hdr;
-        }
-        io_hdr.cmdp = cdb;
-        io_hdr.sbp = sbuff;
-        io_hdr.timeout = immed ? SHORT_TIMEOUT : FORMAT_TIMEOUT;
+        res = sg_ll_format_unit(fd, pinfo, rto_req, 0 /* longlist */,
+                                (!! immed) /* fmtdata */, 0 /* cmplist */,
+                                0 /* dlist_format */,
+                                (immed ? SHORT_TIMEOUT : FORMAT_TIMEOUT),
+                                fmt_hdr, (immed ? sizeof(fmt_hdr) : 0),
+                                1, verbose);
 
-        if (verbose) {
-                fprintf(stderr, "    format cdb: ");
-                for (k = 0; k < 6; ++k)
-                        fprintf(stderr, "%02x ", cdb[k]);
-                fprintf(stderr, "\n");
-        }
-        if ((verbose > 1) && immed) {
-                fprintf(stderr, "    format parameter block\n");
-                dStrHex((const char *)fmt_hdr, FORMAT_HEADER_SIZE, -1);
-        }
-
-        if (ioctl(fd, SG_IO, &io_hdr) < 0) {
-                perror("FORMAT UNIT ioctl error");
-                return -1;
-        }
-        if (verbose > 2)
-                fprintf(stderr, "      duration=%u ms\n", io_hdr.duration);
-        res = sg_err_category3(&io_hdr);
         switch (res) {
-        case SG_LIB_CAT_RECOVERED:
-            sg_chk_n_print3("Format, continuing", &io_hdr, verbose);
-            /* fall through */
-        case SG_LIB_CAT_CLEAN:
+        case 0:
+                break;
+        case SG_LIB_CAT_NOT_READY:
+                fprintf(stderr, "Format command, device not ready\n");
                 break;
         case SG_LIB_CAT_INVALID_OP:
                 fprintf(stderr, "Format command not supported\n");
-                if (verbose > 1)
-                        sg_chk_n_print3("Format", &io_hdr, 1);
-                return -1;
+                break;
         case SG_LIB_CAT_ILLEGAL_REQ:
-                fprintf(stderr, "Format command illegal parameter\n");
-                if (verbose > 1)
-                        sg_chk_n_print3("Format", &io_hdr, 1);
-                return -1;
+                fprintf(stderr, "Format command, illegal parameter\n");
+                break;
+        case SG_LIB_CAT_UNIT_ATTENTION:
+                fprintf(stderr, "Format command, unit attention\n");
+                break;
         default:
-                if (verbose > 1)
-                        sg_chk_n_print3("Format", &io_hdr, 1);
-                return -1;
+                fprintf(stderr, "Format command failed\n");
+                break;
         }
+        if (res)
+                return res;
+
         if (! immed)
                 return 0;
 
         printf("\nFormat has started\n");
         if (early) {
                 if (immed)
-                        printf("Format continuing, use request sense or "
-                               "test unit ready to monitor progress\n");
+                        printf("Format continuing, request sense or test "
+                               "unit ready can be used to monitor progress\n");
                 return 0;
         }
 
         for(;;) {
                 int progress;
-                struct sg_scsi_sense_hdr sshdr;
 
                 sleep(POLL_DURATION_SECS);
-                cdb[0] = TEST_UNIT_READY;       /* draft say REQUEST SENSE */
-                cdb[1] = 0;
-                cdb[2] = 0;
-                cdb[3] = 0;
-                cdb[4] = 0;
-                cdb[5] = 0;
-
-                memset(&io_hdr, 0, sizeof(sg_io_hdr_t));
-                memset(sbuff, 0, MAX_SENSE_SZ);
-
-                io_hdr.interface_id = 'S';
-                io_hdr.dxfer_direction = SG_DXFER_NONE;
-                io_hdr.cmd_len = CDB_SIZE;
-                io_hdr.mx_sb_len = MAX_SENSE_SZ;
-                io_hdr.iovec_count = 0;         /* no scatter gather */
-                io_hdr.dxfer_len = 0;
-                io_hdr.dxferp = NULL;
-                io_hdr.cmdp = cdb;
-                io_hdr.sbp = sbuff;
-                io_hdr.timeout = SHORT_TIMEOUT;
-
-                if (verbose) {
-                        fprintf(stderr, "    test unit ready cdb: ");
-                        for (k = 0; k < 6; ++k)
-                                fprintf(stderr, "%02x ", cdb[k]);
-                        fprintf(stderr, "\n");
-                }
-        
-                if (ioctl(fd, SG_IO, &io_hdr) < 0) {
-                        perror("Test Unit Ready SG_IO ioctl error");
-                        return -1;
-                }
-                if (sg_normalize_sense(&io_hdr, &sshdr)) {
-                        if (sg_get_sense_progress_fld(sbuff,
-                                        io_hdr.sb_len_wr, &progress)) {
-                                printf("Format in progress, %d%% done\n",
-                                        progress * 100 / 65536);
-                                if (verbose > 1)
-                                        sg_print_sense("tur", sbuff,
-                                                       io_hdr.sb_len_wr, 1);
-                                continue;
-                        } else {
-                                sg_print_sense("tur: unexpected sense", sbuff,
-                                               io_hdr.sb_len_wr, verbose);
-                                continue;
-                        }
-                } else
+                progress = -1;
+                res = sg_ll_test_unit_ready_progress(fd, 0, &progress, 0,
+                                                      verbose);
+                if (progress >= 0)
+                        printf("Format in progress, %d%% done\n",
+                                (progress * 100) / 65536);
+                else
                         break;
         }
         printf("FORMAT Complete\n");
@@ -316,13 +193,16 @@ print_read_cap(int fd, int do_16, int verbose)
                         return (int)block_size;
                 }
         }
-        if (SG_LIB_CAT_INVALID_OP == res) 
+        if (SG_LIB_CAT_NOT_READY == res) 
+                fprintf(stderr, "READ CAPACITY (%d): device not ready\n",
+                        (do_16 ? 16 : 10));
+        else if (SG_LIB_CAT_INVALID_OP == res) 
                 fprintf(stderr, "READ CAPACITY (%d) not supported\n",
                         (do_16 ? 16 : 10));
-        if (SG_LIB_CAT_ILLEGAL_REQ == res)
+        else if (SG_LIB_CAT_ILLEGAL_REQ == res)
                 fprintf(stderr, "bad field in READ CAPACITY (%d) "
                         "cdb\n", (do_16 ? 16 : 10));
-        if (verbose)
+        else if (verbose)
                 fprintf(stderr, "READ CAPACITY (%d) failed "
                         "[res=%d]\n", (do_16 ? 16 : 10), res);
         return -1;
@@ -332,7 +212,8 @@ static void usage()
 {
         printf("usage: sg_format [--count=<block count>] [--early] [--format]"
                 " [--help]\n"
-                "                 [--long] [--pinfo] [--resize] [--rto_req]\n"
+                "                 [--long] [--pinfo] [--resize] [--rto_req] "
+                "[--six]\n"
                 "                 [--size=<block size>] [--verbose]"
                 " [--version] [--wait]\n"
                 "                 <scsi_disk>\n"
@@ -355,6 +236,7 @@ static void usage()
                 "value\n"
                 "    --rto_req | -R  set the RTO_REQ bit in format (only valid "
                 "with '--pinfo')\n"
+                "    --six | -6      use 6 byte MODE SENSE/SELECT\n"
                 "    --size=<block size> | -s <block size>\n"
                 "                   only needed to change block size"
                 " (default to\n"
@@ -392,15 +274,16 @@ int main(int argc, char **argv)
         int long_lba = 0;
         int early = 0;
         char device_name[256];
+        char pdt_name[64];
         struct sg_simple_inquiry_resp inq_out;
-        int ret = 1;
+        int ret = 0;
 
         device_name[0] = '\0';
         while (1) {
                 int option_index = 0;
-                char c;
+                int c;
 
-                c = getopt_long(argc, argv, "c:eFhlprRs:vVw",
+                c = getopt_long(argc, argv, "c:eFhlprRs:vVw6",
                                 long_options, &option_index);
                 if (c == -1)
                         break;
@@ -414,7 +297,7 @@ int main(int argc, char **argv)
                                 if (-1 == blk_count) {
                                         fprintf(stderr, "bad argument to "
                                                 "'--count'\n");
-                                        return 1;
+                                        return SG_LIB_SYNTAX_ERROR;
                                 }
                         }
                         break;
@@ -445,7 +328,7 @@ int main(int argc, char **argv)
                         if (blk_size <= 0) {
                                 fprintf(stderr, "bad argument to '--size', "
                                         "want arg > 0)\n");
-                                return 1;
+                                return SG_LIB_SYNTAX_ERROR;
                         }
                         break;
                 case 'v':
@@ -458,9 +341,12 @@ int main(int argc, char **argv)
                 case 'w':
                         fwait = 1;
                         break;
+                case '6':
+                        mode6 = 1;
+                        break;
                 default:
                         usage();
-                        return 1;
+                        return SG_LIB_SYNTAX_ERROR;
                 }
         }
         if (optind < argc) {
@@ -476,29 +362,29 @@ int main(int argc, char **argv)
                         fprintf(stderr, "Unexpected extra argument: %s\n",
                                 argv[optind]);
                 usage();
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
         }
         if ('\0' == device_name[0]) {
                 fprintf(stderr, "no device name given\n");
                 usage();
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
         }
         if (resize) {
                 if (format) {
                         fprintf(stderr, "both '--format' and '--resize'"
                                 "not permitted\n");
                         usage();
-                        return 1;
+                        return SG_LIB_SYNTAX_ERROR;
                 } else if (0 == blk_count) {
                         fprintf(stderr, "'--resize' needs a '--count' (other"
                                 " than 0)\n");
                         usage();
-                        return 1;
+                        return SG_LIB_SYNTAX_ERROR;
                 } else if (0 != blk_size) {
                         fprintf(stderr, "'--resize' not compatible with "
                                 "'--size')\n");
                         usage();
-                        return 1;
+                        return SG_LIB_SYNTAX_ERROR;
                 }
         }
 
@@ -507,21 +393,22 @@ int main(int argc, char **argv)
         ** o verify SCSI device is a disk (get inquiry data first)
         */
 
-        if ((fd = open(device_name, O_RDWR | O_NONBLOCK)) < 0) {
-                char ebuff[128];
-                sprintf(ebuff, "error opening device file: %s", device_name);
-                perror(ebuff);
-                return 1;
+        if ((fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose)) < 0) {
+                fprintf(stderr, "error opening device file: %s: %s\n",
+                        device_name, safe_strerror(-fd));
+                return SG_LIB_FILE_ERROR;
         }
 
         if (sg_simple_inquiry(fd, &inq_out, 1, verbose)) {
                 fprintf(stderr, "%s doesn't respond to a SCSI INQUIRY\n",
                         device_name);
+                ret = SG_LIB_CAT_OTHER;
                 goto out;
         }
         printf("    %.8s  %.16s  %.4s   peripheral_type: %s [0x%x]\n",
                inq_out.vendor, inq_out.product, inq_out.revision,
-               get_ptype_str(inq_out.peripheral_type),
+               sg_get_pdt_str(inq_out.peripheral_type, sizeof(pdt_name),
+                              pdt_name),
                inq_out.peripheral_type);
         if (verbose)
                 printf("      PROTECT=%d\n", !!(inq_out.byte_5 & 1));
@@ -532,6 +419,7 @@ int main(int argc, char **argv)
             (0xe != inq_out.peripheral_type)) {
                 fprintf(stderr, "This format is only defined for disks "
                         "(using SBC-2 or RBC)\n");
+                ret = SG_LIB_CAT_MALFORMED;
                 goto out;
         }
 
@@ -545,11 +433,21 @@ int main(int argc, char **argv)
                                          0 /* current */, mode_page,
                                          0 /* subpage */, dbuff,
                                          MAX_BUFF_SZ, 1, verbose);
+        ret = res;
         if (res) {
-                if (SG_LIB_CAT_INVALID_OP == res)
+                if (SG_LIB_CAT_NOT_READY == res)
+                        fprintf(stderr, "MODE SENSE (%d) command, device "
+                                "not ready\n", (mode6 ? 6 : 10));
+                else if (SG_LIB_CAT_UNIT_ATTENTION == res)
+                        fprintf(stderr, "MODE SENSE (%d) command, unit "
+                                "attention\n", (mode6 ? 6 : 10));
+                else if (SG_LIB_CAT_INVALID_OP == res) {
                         fprintf(stderr, "MODE SENSE (%d) command is not "
                                 "supported\n", (mode6 ? 6 : 10));
-                else if (SG_LIB_CAT_ILLEGAL_REQ == res) {
+                        fprintf(stderr, "    try again %s the '--six' "
+                                "option\n", (mode6 ? "without" : "with"));
+        
+                } else if (SG_LIB_CAT_ILLEGAL_REQ == res) {
                         if (long_lba && (! mode6)) 
                                 fprintf(stderr, "bad field in MODE SENSE "
                                         "(%d) [longlba flag not supported?]"
@@ -666,6 +564,7 @@ int main(int argc, char **argv)
                                 "change number or blocks or block length)\n");
                         fprintf(stderr, "but (single) block descriptor not "
                                 "found in earlier MODE SENSE\n");
+                        ret = SG_LIB_CAT_MALFORMED;
                         goto out;
                 }
                 if (blk_count != 0)  {
@@ -696,8 +595,15 @@ int main(int argc, char **argv)
                 else
                         res = sg_ll_mode_select10(fd, 1 /* PF */, 1 /* SP */,
                                                   dbuff, calc_len, 1, verbose);
+                ret = res;
                 if (res) {
-                        if (SG_LIB_CAT_INVALID_OP == res)
+                        if (SG_LIB_CAT_NOT_READY == res)
+                                fprintf(stderr, "MODE SELECT command, "
+                                        "device not ready\n");
+                        else if (SG_LIB_CAT_UNIT_ATTENTION == res)
+                                fprintf(stderr, "MODE SELECT command, "
+                                        "unit attention\n");
+                        else if (SG_LIB_CAT_INVALID_OP == res)
                                 fprintf(stderr, "MODE SELECT (%d) command is "
                                         "not supported\n", (mode6 ? 6 : 10));
                         else if (SG_LIB_CAT_ILLEGAL_REQ == res)
@@ -706,19 +612,20 @@ int main(int argc, char **argv)
                         else
                                 fprintf(stderr, "MODE SELECT (%d) command "
                                         "failed\n", (mode6 ? 6 : 10));
-                                if (0 == verbose)
-                                        fprintf(stderr, "    try '-v' for "
-                                                "more information\n");
+                        if (0 == verbose)
+                                fprintf(stderr, "    try '-v' for "
+                                        "more information\n");
                         goto out;
                 }
         }
         if (resize) {
-                ret = 0;
                 printf("Resize operation seems to have been successful\n");
                 goto out;
         }
         else if (! format) {
                 res = print_read_cap(fd, do_rcap16, verbose);
+                if (res < 0)
+                        ret = -1;
                 if ((res > 0) && (bd_blk_len > 0) &&
                     (res != (int)bd_blk_len)) {
                         printf("  Warning: mode sense and read capacity "
@@ -728,7 +635,6 @@ int main(int argc, char **argv)
                 }
                 printf("No changes made. To format use '--format'. To "
                        "resize use '--resize'\n");
-                ret = 0;
                 goto out;
         }
 
@@ -744,6 +650,7 @@ int main(int argc, char **argv)
                 sleep(5);
                 res = scsi_format(fd, pinfo, rto_req, ! fwait, early,
                                   verbose);
+                ret = res;
                 if (res) {
                         fprintf(stderr, "FORMAT failed\n");
                         if (0 == verbose)
@@ -753,9 +660,13 @@ int main(int argc, char **argv)
 #else
                 fprintf(stderr, "FORMAT ignored, testing\n");
 #endif
-        ret = 0;
 
 out:
-        close(fd);
-        return ret;
+        res = sg_cmds_close_device(fd);
+        if (res < 0) {
+                fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+                if (0 == ret)
+                        return SG_LIB_FILE_ERROR;
+        }
+        return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
