@@ -4,15 +4,12 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
-#include <errno.h>
-#include <sys/ioctl.h>
-#include <sys/types.h>
-#include "sg_include.h"
+
 #include "sg_lib.h"
 #include "sg_cmds.h"
 
 /* A utility program for the Linux OS SCSI generic ("sg") device driver.
-*  Copyright (C) 2000-2005 D. Gilbert
+*  Copyright (C) 2000-2006 D. Gilbert
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -24,7 +21,7 @@
    
 */
 
-static char * version_str = "1.10 20050904";
+static char * version_str = "1.16 20060623";
 
 #define ME "sg_modes: "
 
@@ -39,39 +36,6 @@ static char * version_str = "1.10 20050904";
 #define EBUFF_SZ 256
 
 
-static const char * scsi_ptype_strs[] = {
-    "disk",                             /* 0x0 */
-    "tape",
-    "printer",
-    "processor",
-    "write once optical disk",
-    "cd/dvd",
-    "scanner",
-    "optical memory device",
-    "medium changer",                   /* 0x8 */
-    "communications",
-    "graphics [0xa]",
-    "graphics [0xb]",
-    "storage array controller",
-    "enclosure services device",
-    "simplified direct access device",
-    "optical card reader/writer device",
-    "bridge controller commands",       /* 0x10 */
-    "object storage device",
-    "automation/drive interface",
-    "0x13", "0x14", "0x15", "0x16", "0x17", "0x18",
-    "0x19", "0x1a", "0x1b", "0x1c", "0x1d",
-    "well known logical unit",
-    "no physical device on this lu",
-};
-
-static const char * get_ptype_str(int scsi_ptype)
-{
-    int num = sizeof(scsi_ptype_strs) / sizeof(scsi_ptype_strs[0]);
-
-    return (scsi_ptype < num) ? scsi_ptype_strs[scsi_ptype] : "";
-}
-
 static const char * transport_proto_arr[] =
 {
     "Fibre Channel (FCP-2)",
@@ -94,7 +58,7 @@ struct page_code_desc {
 };
 
 static struct page_code_desc pc_desc_common[] = {
-    {0x0, 0x0, "Unit Attention condition [vendor: page format optional]"},
+    {0x0, 0x0, "Unit Attention condition [vendor specific format]"},
     {0x2, 0x0, "Disconnect-Reconnect"},
     {0x9, 0x0, "Peripheral device (obsolete)"},
     {0xa, 0x0, "Control"},
@@ -127,8 +91,8 @@ static struct page_code_desc pc_desc_disk[] = {
 
 static struct page_code_desc pc_desc_tape[] = {
     {0xf, 0x0, "Data Compression"},
-    {0x10, 0x0, "Device configation"},
-    {0x10, 0x1, "Device configation extension"},
+    {0x10, 0x0, "Device configuration"},
+    {0x10, 0x1, "Device configuration extension"},
     {0x11, 0x0, "Medium Partition [1]"},
     {0x12, 0x0, "Medium Partition [2]"},
     {0x13, 0x0, "Medium Partition [3]"},
@@ -236,6 +200,7 @@ static struct page_code_desc pc_desc_t_sas[] = {
     {0x18, 0x0, "LU SSP, short format"},
     {0x19, 0x0, "Port SSP, short format"},
     {0x19, 0x1, "Port SSP, phy control and discover"},
+    {0x19, 0x2, "Port SSP, shared"},
 };
 
 static struct page_code_desc pc_desc_t_adt[] = {
@@ -429,6 +394,47 @@ static void list_page_codes(int scsi_ptype, int inq_byte6, int t_proto)
     }
 }
 
+static int examine_pages(int sg_fd, int do_mode6, int inq_pdt, int inq_byte6,
+                         int verbose)
+{
+    int k, res, header;
+    unsigned char rbuf[4];
+    const char * cp;
+
+    for (header = 0, k = 0; k < 0x3f; ++k) {
+        if (do_mode6) {
+            res = sg_ll_mode_sense6(sg_fd, 0, 0, k, 0,
+                                    rbuf, sizeof(rbuf), 0, verbose);
+            if (SG_LIB_CAT_INVALID_OP == res) {
+                fprintf(stderr, ">>>>>> try again without the '-6' "
+                        "switch for a 10 byte MODE SENSE command\n");
+                return res;
+            }
+        } else {
+            res = sg_ll_mode_sense10(sg_fd, 0, 0, 0, k,
+                                     0, rbuf, sizeof(rbuf), 1, verbose);
+            if (SG_LIB_CAT_INVALID_OP == res) {
+                fprintf(stderr, ">>>>>> try again with a '-6' "
+                        "switch for a 6 byte MODE SENSE command\n");
+                return res;
+            }
+        }
+        if (0 == res) {
+            if (0 == header) {
+                printf("Discovered mode pages:\n");
+                header = 1;
+            }
+            cp = find_page_code_desc(k, 0, inq_pdt, inq_byte6, -1);
+            if (cp)
+                printf("    %s\n", cp);
+            else
+                printf("    [0x%x]\n", k);
+        } else if (SG_LIB_CAT_NOT_READY == res)
+            fprintf(stderr, "MODE SENSE failed, device not ready\n");
+    }
+    return res;
+}
+
 static const char * pg_control_str_arr[] = {
     "current",
     "changeable",
@@ -438,15 +444,16 @@ static const char * pg_control_str_arr[] = {
 
 static void usage()
 {
-    printf("Usage:  sg_modes [-a] [-A] [-c=<page_control] [-d] [-D] [-f] [-h] "
-           "[-H] [-l]\n\t\t"
-           " [-L] [-p=<page_number>[,<sub_page_code>]] [-r]"
+    printf("Usage:  sg_modes [-a] [-A] [-c=<page_control] [-d] [-D] [-f] "
+           "[-e] [-h] [-H]\n\t\t"
+           " [-l] [-L] [-p=<page_number>[,<sub_page_code>]] [-r]"
            "\n\t\t [-subp=<sub_page_code>] [-v] [-V] [-6] [<scsi_device>]\n"
            " where -a   get all mode pages supported by device\n"
            "       -A   get all mode pages and subpages supported by device\n"
            "       -c=<page_control> page control (def: 0 [current],"
            " 1 [changeable],\n            2 [default], 3 [saved])\n"
            "       -d   disable block descriptors (DBD field in cdb)\n"
+           "       -e   examine pages # 0 through to 0x3e, note if found\n"
            "       -D   disable block descriptor output\n"
            "       -f   be flexible, cope with MODE SENSE 6/10 response "
            "mixup\n");
@@ -487,6 +494,7 @@ int main(int argc, char * argv[])
     int do_all_sub = 0;
     int do_dbd = 0;
     int no_desc_out = 0;
+    int do_examine = 0;
     int flexible = 0;
     int do_hex = 0;
     int do_llbaa = 0;
@@ -494,12 +502,13 @@ int main(int argc, char * argv[])
     int do_list = 0;
     int do_raw = 0;
     int do_verbose = 0;
-    int oflags = O_RDONLY | O_NONBLOCK;
+    int ret = 0;
     int density_code_off, t_proto, inq_pdt, inq_byte6, resp_mode6;
     int num_ua_pages, plen, jmp_out;
     unsigned char * ucp;
     unsigned char uc;
     struct sg_simple_inquiry_resp inq_out;
+    char pdt_name[64];
 
     for (k = 1; k < argc; ++k) {
         cp = argv[k];
@@ -525,6 +534,9 @@ int main(int argc, char * argv[])
                 case 'D':
                     no_desc_out = 1;
                     break;
+                case 'e':
+                    do_examine = 1;
+                    break;
                 case 'f':
                     flexible = 1;
                     break;
@@ -549,7 +561,7 @@ int main(int argc, char * argv[])
                     exit(0);
                 case '?':
                     usage();
-                    return 1;
+                    return 0;
                 default:
                     jmp_out = 1;
                     break;
@@ -564,7 +576,7 @@ int main(int argc, char * argv[])
                 if ((1 != num) || (u > 3)) {
                     fprintf(stderr, "Bad page control after 'c=' option\n");
                     usage();
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
                 pc = u;
             } else if (0 == strncmp("p=", cp, 2)) {
@@ -574,7 +586,7 @@ int main(int argc, char * argv[])
                         fprintf(stderr, "Bad page code value after 'p=' "
                                 "option\n");
                         usage();
-                        return 1;
+                        return SG_LIB_SYNTAX_ERROR;
                     }
                     pg_code = u;
                 } else if (2 == sscanf(cp + 2, "%x,%x", &u, &uu)) {
@@ -582,7 +594,7 @@ int main(int argc, char * argv[])
                         fprintf(stderr, "Bad sub page code value after 'p=' "
                                 "option\n");
                         usage();
-                        return 1;
+                        return SG_LIB_SYNTAX_ERROR;
                     }
                     pg_code = u;
                     sub_pg_code = uu;
@@ -591,7 +603,7 @@ int main(int argc, char * argv[])
                     fprintf(stderr, "Bad page code, subpage code sequence "
                             "after 'p=' option\n");
                     usage();
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
             } else if (0 == strncmp("subp=", cp, 5)) {
                 num = sscanf(cp + 5, "%x", &u);
@@ -599,7 +611,7 @@ int main(int argc, char * argv[])
                     fprintf(stderr, "Bad sub page code after 'subp=' "
                             "option\n");
                     usage();
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
                 sub_pg_code = u;
                 sub_pg_code_set = 1;
@@ -608,7 +620,7 @@ int main(int argc, char * argv[])
             } else if (jmp_out) {
                 fprintf(stderr, "Unrecognized option: %s\n", cp);
                 usage();
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == file_name)
             file_name = cp;
@@ -616,7 +628,7 @@ int main(int argc, char * argv[])
             fprintf(stderr, "too many arguments, got: %s, not expecting: "
                     "%s\n", file_name, cp);
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     
@@ -627,7 +639,7 @@ int main(int argc, char * argv[])
                 list_page_codes(0, 0, -1);
             } else {
                 printf("    peripheral device type: %s\n",
-                       get_ptype_str(pg_code));
+                       sg_get_pdt_str(pg_code, sizeof(pdt_name), pdt_name));
                 if (sub_pg_code_set)
                     list_page_codes(pg_code, 0, sub_pg_code);
                 else
@@ -637,7 +649,12 @@ int main(int argc, char * argv[])
         }
         fprintf(stderr, "No <scsi_device> argument given\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
+    }
+
+    if (do_examine && (pg_code >= 0)) {
+        fprintf(stderr, "can't give '-e' and a page number\n");
+        return SG_LIB_SYNTAX_ERROR;
     }
 
     /* The 6 bytes command only allows up to 252 bytes of response data */
@@ -645,39 +662,43 @@ int main(int argc, char * argv[])
         if (do_llbaa) {
             fprintf(stderr, "LLBAA not defined for MODE SENSE 6, try "
                     "without '-L'\n");
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
         rsp_buff_size = 252;
     }
     /* If no pages or list selected than treat as 'a' */
-    if (! ((pg_code >= 0) || do_all || do_list))
+    if (! ((pg_code >= 0) || do_all || do_list || do_examine))
         do_all = 1;
 
-    if ((sg_fd = open(file_name, oflags)) < 0) {
-        snprintf(ebuff, EBUFF_SZ, ME "error opening file: %s", file_name);
-        perror(ebuff);
-        return 1;
+    if ((sg_fd = sg_cmds_open_device(file_name, 1 /* ro */, do_verbose)) < 0) {
+        fprintf(stderr, ME "error opening file: %s: %s\n", file_name,
+                safe_strerror(-sg_fd));
+        return SG_LIB_FILE_ERROR;
     }
 
     if (sg_simple_inquiry(sg_fd, &inq_out, 1, do_verbose)) {
         fprintf(stderr, ME "%s doesn't respond to a SCSI INQUIRY\n",
                 file_name);
-        close(sg_fd);
-        return 1;
+        sg_cmds_close_device(sg_fd);
+        return SG_LIB_CAT_OTHER;
     }
     inq_pdt = inq_out.peripheral_type;
     inq_byte6 = inq_out.byte_6;
     if (0 == do_raw)
         printf("    %.8s  %.16s  %.4s   peripheral_type: %s [0x%x]\n", 
                inq_out.vendor, inq_out.product, inq_out.revision,
-               get_ptype_str(inq_pdt), inq_pdt);
+               sg_get_pdt_str(inq_pdt, sizeof(pdt_name), pdt_name), inq_pdt);
     if (do_list) {
         if (sub_pg_code_set)
-            list_page_codes(inq_pdt, inq_pdt, sub_pg_code);
+            list_page_codes(inq_pdt, inq_byte6, sub_pg_code);
         else
-            list_page_codes(inq_pdt, inq_pdt, -1);
+            list_page_codes(inq_pdt, inq_byte6, -1);
         return 0;
     }
+    if (do_examine) {
+        ret = examine_pages(sg_fd, do_mode6, inq_pdt, inq_byte6, do_verbose);
+        goto finish;
+    }
     if (PG_CODE_ALL == pg_code)
         do_all = 1;
     else if (do_all)
@@ -689,12 +710,12 @@ int main(int argc, char * argv[])
         if (do_all) {
             fprintf(stderr, "'-r' requires a given (sub)page (not all)\n");
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
         if (do_hex) {
             fprintf(stderr, "'-r' and '-h' clash");
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
 
@@ -706,7 +727,7 @@ int main(int argc, char * argv[])
             fprintf(stderr, ">>>>>> try again without the '-6' "
                     "switch for a 10 byte MODE SENSE command\n");
         else if (SG_LIB_CAT_ILLEGAL_REQ == res)
-            fprintf(stderr, "bad field in cdb (perhaps subpages "
+            fprintf(stderr, "invalid field in cdb (perhaps subpages "
                     "or page control (PC) not supported)\n");
     } else {
         res = sg_ll_mode_sense10(sg_fd, do_llbaa, do_dbd, pc, pg_code,
@@ -716,12 +737,21 @@ int main(int argc, char * argv[])
             fprintf(stderr, ">>>>>> try again with a '-6' "
                     "switch for a 6 byte MODE SENSE command\n");
         else if (SG_LIB_CAT_ILLEGAL_REQ == res)
-            fprintf(stderr, "bad field in cdb (perhaps subpages "
+            fprintf(stderr, "invalid field in cdb (perhaps subpages "
                     "or page control (PC) not supported)\n");
     }
+    if (SG_LIB_CAT_ILLEGAL_REQ == res)
+        fprintf(stderr, "invalid field in cdb (perhaps subpages "
+                "or page control (PC) not supported)\n");
+    else if (SG_LIB_CAT_NOT_READY == res)
+        fprintf(stderr, "device not ready\n");
+    else if (SG_LIB_CAT_UNIT_ATTENTION == res)
+        fprintf(stderr, "unit attention\n");
+    ret = res;
     if (0 == res) {
         int medium_type, specific, headerlen;
 
+        ret = 0;
         resp_mode6 = do_mode6;
         if (flexible) {
             num = rsp_buff[0];
@@ -770,7 +800,7 @@ int main(int argc, char * argv[])
             len = (len < md_len) ? len : md_len;
             for (k = 0; k < len; ++k)
                 printf("%02x\n", ucp[k]);
-            close(sg_fd);
+            sg_cmds_close_device(sg_fd);
             return 0;
         }
         if (do_hex)
@@ -882,6 +912,7 @@ int main(int argc, char * argv[])
             md_len -= len;
         }
     }
-    close(sg_fd);
-    return 0;
+finish:
+    sg_cmds_close_device(sg_fd);
+    return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
