@@ -8,12 +8,12 @@
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include "sg_include.h"
 #include "sg_lib.h"
 #include "sg_cmds.h"
+#include "sg_io_linux.h"
 
 /* A utility program for the Linux OS SCSI subsystem.
-*  Copyright (C) 2004-2005 D. Gilbert
+*  Copyright (C) 2004-2006 D. Gilbert
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -24,7 +24,7 @@
 
 */
 
-static char * version_str = "0.21 20050904";
+static char * version_str = "0.24 20060623";
 
 
 #define SENSE_BUFF_LEN 32       /* Arbitrary, could be larger */
@@ -92,7 +92,7 @@ static unsigned char dummy_rsmft_r0 = 0xff;
 
 
 /* Report Supported Operation Codes */
-/* Returns 0 when successful, else -1 */
+/* Returns 0 when successful */
 static int do_rsoc(int sg_fd, int rep_opts, int rq_opcode, int rq_servact, 
                   void * resp, int mx_resp_len, int noisy, int verbose)
 {
@@ -143,7 +143,7 @@ static int do_rsoc(int sg_fd, int rep_opts, int rq_opcode, int rq_servact,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         sg_chk_n_print3("Report supported operation codes", &io_hdr,
-                        verbose);
+                        verbose > 1);
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         return 0;
@@ -159,14 +159,14 @@ static int do_rsoc(int sg_fd, int rep_opts, int rq_opcode, int rq_servact,
             else
                 snprintf(ebuff, EBUFF_SZ, "RSOC error, rq_opcode=0x%x, "
                          "rq_sa=0x%x ", rq_opcode, rq_servact);
-            sg_chk_n_print3(ebuff, &io_hdr, verbose);
+            sg_chk_n_print3(ebuff, &io_hdr, verbose > 1);
         }
-        return -1;
+        return res;
     }
 }
 
 /* Report Supported Task Management Function */
-/* Returns 0 when successful, else -1 */
+/* Returns 0 when successful */
 static int do_rstmf(int sg_fd, void * resp, int mx_resp_len, int noisy,
                     int verbose)
 {
@@ -208,7 +208,7 @@ static int do_rstmf(int sg_fd, void * resp, int mx_resp_len, int noisy,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         sg_chk_n_print3("Report supported task management fns", &io_hdr,
-                        verbose);
+                        verbose > 1);
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         return 0;
@@ -216,9 +216,9 @@ static int do_rstmf(int sg_fd, void * resp, int mx_resp_len, int noisy,
         if (noisy | verbose) {
             char ebuff[EBUFF_SZ];
             snprintf(ebuff, EBUFF_SZ, "RSTMF error ");
-            sg_chk_n_print3(ebuff, &io_hdr, verbose);
+            sg_chk_n_print3(ebuff, &io_hdr, verbose > 1);
         }
-        return -1;
+        return res;
     }
 }
 
@@ -241,39 +241,6 @@ static void usage()
             "management\nfunctions) SCSI command\n");
 }
 
-static const char * scsi_ptype_strs[] = {
-    /* 0 */ "disk",
-    "tape",
-    "printer",
-    "processor",
-    "write once optical disk",
-    /* 5 */ "cd/dvd",
-    "scanner",
-    "optical memory device",
-    "medium changer",
-    "communications",
-    /* 0xa */ "graphics [0xa]",
-    "graphics [0xb]",
-    "storage array controller",
-    "enclosure services device",
-    "simplified direct access device",
-    "optical card reader/writer device",
-    /* 0x10 */ "bridge controller commands",
-    "object based storage",
-    "automation/driver interface",
-    "0x13", "0x14", "0x15", "0x16", "0x17", "0x18",
-    "0x19", "0x1a", "0x1b", "0x1c", "0x1d",
-    "well known logical unit",
-    "no physical device on this lu",
-};
-
-static const char * get_ptype_str(int scsi_ptype)
-{
-    int num = sizeof(scsi_ptype_strs) / sizeof(scsi_ptype_strs[0]);
-
-    return (scsi_ptype < num) ? scsi_ptype_strs[scsi_ptype] : "";
-}
-
 /* returns -1 when left < right, 0 when left == right, else returns 1 */
 int opcode_num_compare(const void * left, const void * right)
 {
@@ -392,7 +359,7 @@ void list_all_codes(unsigned char * rsoc_buff, int rsoc_len, int unsorted,
 
 int main(int argc, char * argv[])
 {
-    int sg_fd, k, num, cd_len, plen, jmp_out;
+    int sg_fd, k, num, cd_len, plen, jmp_out, res;
     const char * file_name = 0;
     char ebuff[EBUFF_SZ];
     unsigned char rsoc_buff[MX_ALLOC_LEN];
@@ -405,8 +372,8 @@ int main(int argc, char * argv[])
     int do_unsorted = 0;
     int do_taskman = 0;
     int rep_opts = 0;
-    int ret = 0;
     const char * cp;
+    char buff[48];
     struct sg_simple_inquiry_resp inq_resp;
 
     for (k = 1; k < argc; ++k) {
@@ -435,7 +402,7 @@ int main(int argc, char * argv[])
                 case 'h':
                 case '?':
                     usage();
-                    return 1;
+                    return 0;
                 default:
                     jmp_out = 1;
                     break;
@@ -450,19 +417,19 @@ int main(int argc, char * argv[])
                 if ((1 != num) || (do_opcode > 255)) {
                     fprintf(stderr, "Bad number after 'o=' option\n");
                     usage();
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
             } else if (0 == strncmp("s=", cp, 2)) {
                 num = sscanf(cp + 2, "%x", (unsigned int *)&do_servact);
                 if (1 != num) {
                     fprintf(stderr, "Bad number after 's=' option\n");
                     usage();
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
             } else if (jmp_out) {
                 fprintf(stderr, "Unrecognized option: %s\n", cp);
                 usage();
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == file_name)
             file_name = cp;
@@ -470,19 +437,19 @@ int main(int argc, char * argv[])
             fprintf(stderr, "too many arguments, got: %s, not expecting: "
                     "%s\n", file_name, cp);
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     
     if (0 == file_name) {
         fprintf(stderr, "No <scsi_device> argument given\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     if ((-1 != do_servact) && (-1 == do_opcode)) {
         fprintf(stderr, "When '-s' is chosen, so must '-o' be chosen\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     if (do_unsorted && do_alpha)
         fprintf(stderr, "warning: unsorted ('-u') and alpha ('-a') options "
@@ -497,20 +464,20 @@ int main(int argc, char * argv[])
         snprintf(ebuff, EBUFF_SZ, "sg_opcodes: error opening file: %s (ro)",
                  file_name);
         perror(ebuff);
-        return 1;
+        return SG_LIB_FILE_ERROR;
     }
     if (0 == sg_simple_inquiry(sg_fd, &inq_resp, 1, do_verbose)) {
         printf("  %.8s  %.16s  %.4s\n", inq_resp.vendor, inq_resp.product,
                inq_resp.revision);
         peri_type = inq_resp.peripheral_type;
-        cp = get_ptype_str(peri_type);
+        cp = sg_get_pdt_str(peri_type, sizeof(buff), buff);
         if (strlen(cp) > 0)
             printf("  Peripheral device type: %s\n", cp);
         else
             printf("  Peripheral device type: 0x%x\n", peri_type);
     } else {
         printf("sg_opcodes: %s doesn't respond to a SCSI INQUIRY\n", file_name);
-        return 1;
+        return SG_LIB_CAT_OTHER;
     }
     close(sg_fd);
 #ifndef TEST_CODE
@@ -521,7 +488,7 @@ int main(int argc, char * argv[])
         else
             printf("'Report supported operation codes' command not "
                    "supported for CD/DVD devices\n");
-        return 1;
+        return SG_LIB_CAT_OTHER;
     }
 #endif
 
@@ -529,20 +496,26 @@ int main(int argc, char * argv[])
         snprintf(ebuff, EBUFF_SZ, "sg_opcodes: error opening file: %s (rw)",
                  file_name);
         perror(ebuff);
-        return 1;
+        return SG_LIB_FILE_ERROR;
     }
     if (do_opcode >= 0)
         rep_opts = ((do_servact >= 0) ? 2 : 1);
     memset(rsoc_buff, 0, sizeof(rsoc_buff));
 #ifndef TEST_CODE
     if (do_taskman) {
-        if (0 != do_rstmf(sg_fd, rsoc_buff, sizeof(rsoc_buff), 1,
-                          do_verbose))
-            return 1;
+        res = do_rstmf(sg_fd, rsoc_buff, sizeof(rsoc_buff), 0,
+                       do_verbose);
+        if (res) {
+            fprintf(stderr, "Report supported task management functions failed\n");
+            return res;
+        }
     } else {
-        if (0 != do_rsoc(sg_fd, rep_opts, do_opcode, do_servact, rsoc_buff,
-                         sizeof(rsoc_buff), 1, do_verbose))
-            return 1;
+        res = do_rsoc(sg_fd, rep_opts, do_opcode, do_servact, rsoc_buff,
+                      sizeof(rsoc_buff), 0, do_verbose);
+        if (res) {
+            fprintf(stderr, "Report supported operation codes failed\n");
+            return res;
+        }
     }
 #else
     if (do_taskman)
@@ -609,5 +582,5 @@ int main(int argc, char * argv[])
         }
     }
     close(sg_fd);
-    return ret;
+    return 0;
 }
