@@ -5,16 +5,12 @@
 #include <string.h>
 #include <ctype.h>
 #include <getopt.h>
-#include <errno.h>
-#include <sys/ioctl.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include "sg_include.h"
+
 #include "sg_lib.h"
 #include "sg_cmds.h"
 
 /* A utility program for the Linux OS SCSI subsystem.
-*  Copyright (C) 2004-2005 D. Gilbert
+*  Copyright (C) 2004-2006 D. Gilbert
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -24,19 +20,13 @@
 
 */
 
-static char * version_str = "0.23 20050808";
+static char * version_str = "0.26 20060623";
 
 
-#define SENSE_BUFF_LEN 32       /* Arbitrary, could be larger */
-#define DEF_TIMEOUT 60000       /* 60,000 millisecs == 60 seconds */
-
-#define SG_PERSISTENT_IN 0x5e
-#define SG_PERSISTENT_OUT 0x5f
 #define PRIN_RKEY_SA     0x0
 #define PRIN_RRES_SA     0x1
 #define PRIN_RCAP_SA     0x2
 #define PRIN_RFSTAT_SA   0x3
-#define PRINOUT_CMD_LEN 10
 #define PROUT_REG_SA     0x0
 #define PROUT_RES_SA     0x1
 #define PROUT_REL_SA     0x2
@@ -47,8 +37,6 @@ static char * version_str = "0.23 20050808";
 #define PROUT_REG_MOVE_SA 0x7
 #define MX_ALLOC_LEN 8192
 
-#define EBUFF_SZ 256
-
 
 static struct option long_options[] = {
         {"clear", 0, 0, 'C'},
@@ -112,131 +100,11 @@ static const int num_prout_sa_strs = sizeof(prout_sa_strs) /
                                      sizeof(prout_sa_strs[0]);
 
 
-/* Returns 0 when successful, else -1 */
-static int do_prin(int sg_fd, int rq_servact, void * resp, int mx_resp_len,
-                   int noisy, int verbose)
-{
-    int res, k;
-    unsigned char prinCmdBlk[PRINOUT_CMD_LEN] = {SG_PERSISTENT_IN, 0, 0, 0,
-                                                 0, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
-    struct sg_io_hdr io_hdr;
-
-    if (rq_servact > 0) {
-        prinCmdBlk[1] = (unsigned char)(rq_servact & 0x1f);
-
-    }
-    prinCmdBlk[7] = (unsigned char)((mx_resp_len >> 8) & 0xff);
-    prinCmdBlk[8] = (unsigned char)(mx_resp_len & 0xff);
-
-    if (verbose) {
-        fprintf(stderr, "    Persistent Reservation In cmd: ");
-        for (k = 0; k < PRINOUT_CMD_LEN; ++k)
-            fprintf(stderr, "%02x ", prinCmdBlk[k]);
-        fprintf(stderr, "\n");
-    }
-    memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
-    io_hdr.interface_id = 'S';
-    io_hdr.cmd_len = sizeof(prinCmdBlk);
-    io_hdr.mx_sb_len = sizeof(sense_b);
-    io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
-    io_hdr.dxfer_len = mx_resp_len;
-    io_hdr.dxferp = resp;
-    io_hdr.cmdp = prinCmdBlk;
-    io_hdr.sbp = sense_b;
-    io_hdr.timeout = DEF_TIMEOUT;
-
-    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
-        perror("SG_IO (PR In) error");
-        return -1;
-    }
-    res = sg_err_category3(&io_hdr);
-    switch (res) {
-    case SG_LIB_CAT_RECOVERED:
-        sg_chk_n_print3("PRIN, continuing", &io_hdr, verbose);
-        /* fall through */
-    case SG_LIB_CAT_CLEAN:
-        return 0;
-    default:
-        if (noisy || verbose) {
-            char ebuff[EBUFF_SZ];
-            snprintf(ebuff, EBUFF_SZ, "PRIN error, service_action: %s",
-                     ((rq_servact < num_prin_sa_strs) ? 
-                        prin_sa_strs[rq_servact] : "??"));
-            sg_chk_n_print3(ebuff, &io_hdr, verbose);
-        }
-        return -1;
-    }
-}
-
-/* Returns 0 when successful, else -1 */
-static int do_prout(int sg_fd, int rq_servact, int rq_scope, 
-                    unsigned int rq_type, void * paramp, int param_len,
-                    int noisy, int verbose)
-{
-    int res, k;
-    unsigned char proutCmdBlk[PRINOUT_CMD_LEN] = {SG_PERSISTENT_OUT, 0, 0, 0,
-                                                  0, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
-    struct sg_io_hdr io_hdr;
-
-    if (rq_servact > 0) {
-        proutCmdBlk[1] = (unsigned char)(rq_servact & 0x1f);
-
-    }
-    proutCmdBlk[2] = (((rq_scope & 0xf) << 4) | (rq_type & 0xf));
-    proutCmdBlk[7] = (unsigned char)((param_len >> 8) & 0xff);
-    proutCmdBlk[8] = (unsigned char)(param_len & 0xff);
-
-    if (verbose) {
-        fprintf(stderr, "    Persistent Reservation Out cmd: ");
-        for (k = 0; k < PRINOUT_CMD_LEN; ++k)
-            fprintf(stderr, "%02x ", proutCmdBlk[k]);
-        fprintf(stderr, "\n");
-        if (verbose > 1) {
-            fprintf(stderr, "    Persistent Reservation Out parameters:\n");
-            dStrHex(paramp, param_len, 0);
-        }
-    }
-    memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
-    io_hdr.interface_id = 'S';
-    io_hdr.cmd_len = sizeof(proutCmdBlk);
-    io_hdr.mx_sb_len = sizeof(sense_b);
-    io_hdr.dxfer_direction = SG_DXFER_TO_DEV;
-    io_hdr.dxfer_len = param_len;
-    io_hdr.dxferp = paramp;
-    io_hdr.cmdp = proutCmdBlk;
-    io_hdr.sbp = sense_b;
-    io_hdr.timeout = DEF_TIMEOUT;
-
-    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
-        perror("SG_IO (PR Out) error");
-        return -1;
-    }
-    res = sg_err_category3(&io_hdr);
-    switch (res) {
-    case SG_LIB_CAT_RECOVERED:
-        sg_chk_n_print3("PROUT, continuing", &io_hdr, verbose);
-        /* fall through */
-    case SG_LIB_CAT_CLEAN:
-        return 0;
-    default:
-        if (noisy) {
-            char ebuff[EBUFF_SZ];
-            snprintf(ebuff, EBUFF_SZ, "PROUT error, service_action: %s",
-                     ((rq_servact < num_prout_sa_strs) ? 
-                        prout_sa_strs[rq_servact] : "??"));
-            sg_chk_n_print3(ebuff, &io_hdr, verbose);
-        }
-        return -1;
-    }
-}
-
 static void usage()
 {
     fprintf(stderr,
             "Usage: 'sg_persist [<options>] [<scsi_device>]\n"
-            " where Persistent Reservation (PR) <options> include:\n"
+            " where Persistent Reserve (PR) <options> include:\n"
             "       --clear|-C             PR Out: Clear\n"
             "       --device=<scsi_device> device to query or change\n"
             "       -d <scsi_device>       device to query or change "
@@ -277,40 +145,7 @@ static void usage()
             "       --verbose|-v   output additional debug information\n"
             "       --version|-V   output version string\n"
             "       -?   output this usage message\n\n"
-            "Performs a PERSISTENT RESERVATION (IN or OUT) SCSI command\n");
-}
-
-static const char * scsi_ptype_strs[] = {
-    /* 0 */ "disk",
-    "tape",
-    "printer",
-    "processor",
-    "write once optical disk",
-    /* 5 */ "cd/dvd",
-    "scanner",
-    "optical memory device",
-    "medium changer",
-    "communications",
-    /* 0xa */ "graphics",
-    "graphics",
-    "storage array controller",
-    "enclosure services device",
-    "simplified direct access device",
-    "optical card reader/writer device",
-    /* 0x10 */ "bridging expander",
-    "object based storage",
-    "automation/driver interface",
-    "0x13", "0x14", "0x15", "0x16", "0x17", "0x18",
-    "0x19", "0x1a", "0x1b", "0x1c", "0x1d",
-    "well known logical unit",
-    "no physical device on this lu",
-};
-
-static const char * get_ptype_str(int scsi_ptype)
-{
-    int num = sizeof(scsi_ptype_strs) / sizeof(scsi_ptype_strs[0]);
-
-    return (scsi_ptype < num) ? scsi_ptype_strs[scsi_ptype] : "";
+            "Performs a PERSISTENT RESERVE (IN or OUT) SCSI command\n");
 }
 
 static const char * pr_type_strs[] = {
@@ -433,15 +268,25 @@ static void decode_transport_id(const char * leadin, unsigned char * ucp,
 
 static int prin_work(int sg_fd, int prin_sa, int do_verbose, int do_hex)
 {
-    int k, j, num, add_len, add_desc_len, rel_pt_addr;
+    int k, j, num, res, add_len, add_desc_len, rel_pt_addr;
     unsigned int pr_gen;
     unsigned long long ull;
     unsigned char * ucp;
     unsigned char pr_buff[MX_ALLOC_LEN];
 
     memset(pr_buff, 0, sizeof(pr_buff));
-    if (0 != do_prin(sg_fd, prin_sa, pr_buff, 
-                     sizeof(pr_buff), 1, do_verbose)) {
+    res = sg_ll_persistent_reserve_in(sg_fd, prin_sa, pr_buff,
+                                      sizeof(pr_buff), 1, do_verbose);
+    if (res) {
+       if (SG_LIB_CAT_INVALID_OP == res)
+            fprintf(stderr, "Persistent reserve in, command not "
+                    "supported\n");
+        else if (SG_LIB_CAT_ILLEGAL_REQ == res)
+            fprintf(stderr, "Persistent reserve in: bad field in cdb\n");
+        else if (SG_LIB_CAT_UNIT_ATTENTION == res)
+            fprintf(stderr, "Persistent reserve in: unit attention\n");
+        else
+            fprintf(stderr, "Persistent reserve in, command failed\n");
         return 1;
     }
     if (PRIN_RCAP_SA == prin_sa) {
@@ -464,6 +309,7 @@ static int prin_work(int sg_fd, int prin_sa, int do_verbose, int do_hex)
                    !!(pr_buff[2] & 0x1));
             printf("  Type Mask Valid(TMV): %d\n",
                    !!(pr_buff[3] & 0x80));
+            printf("  Allow commands: %d\n", (pr_buff[3] >> 4) & 0x7);
             printf("  Persist Through Power Loss active(PTPL_A): %d\n",
                    !!(pr_buff[3] & 0x1));
             if (pr_buff[3] & 0x80) {
@@ -588,7 +434,7 @@ static int prout_work(int sg_fd, int prout_sa, unsigned int prout_type,
                       int param_aptpl, unsigned char * transportidp,
                       int transportid_len, int do_verbose)
 {
-    int j, len;
+    int j, len, res;
     unsigned char pr_buff[MX_ALLOC_LEN];
 
     memset(pr_buff, 0, sizeof(pr_buff));
@@ -614,8 +460,17 @@ static int prout_work(int sg_fd, int prout_sa, unsigned int prout_type,
         pr_buff[26] = (unsigned char)((transportid_len >> 8) & 0xff);
         pr_buff[27] = (unsigned char)(transportid_len & 0xff);
     }
-    if (0 != do_prout(sg_fd, prout_sa, 0, prout_type, pr_buff,
-                      len, 1, do_verbose)) {
+    res = sg_ll_persistent_reserve_out(sg_fd, prout_sa, 0, prout_type,
+                                       pr_buff, len, 1, do_verbose);
+    if (res) {
+       if (SG_LIB_CAT_INVALID_OP == res)
+            fprintf(stderr, "Persistent reserve out, command not supported\n");
+        else if (SG_LIB_CAT_ILLEGAL_REQ == res)
+            fprintf(stderr, "Persistent reserve out, bad field in cdb\n");
+        else if (SG_LIB_CAT_UNIT_ATTENTION == res)
+            fprintf(stderr, "Persistent reserve out, unit attention\n");
+        else
+            fprintf(stderr, "Persistent reserve out, command failed\n");
         return 1;
     } else if (do_verbose) {
         char buff[64];
@@ -624,7 +479,7 @@ static int prout_work(int sg_fd, int prout_sa, unsigned int prout_type,
             strncpy(buff, prout_sa_strs[prout_sa], sizeof(buff));
         else
             snprintf(buff, sizeof(buff), "service action=0x%x", prout_sa);
-        fprintf(stderr, "Persistent Reservation Out command (%s) "
+        fprintf(stderr, "Persistent Reserve Out command (%s) "
                 "successful\n", buff);
     }
     return 0;
@@ -637,7 +492,7 @@ static int prout_rmove_work(int sg_fd, unsigned int prout_type,
                       unsigned char * transportidp, int transportid_len,
                       int do_verbose)
 {
-    int j, len;
+    int j, len, res;
     unsigned char pr_buff[MX_ALLOC_LEN];
 
     memset(pr_buff, 0, sizeof(pr_buff));
@@ -664,11 +519,21 @@ static int prout_rmove_work(int sg_fd, unsigned int prout_type,
         pr_buff[22] = (unsigned char)((transportid_len >> 8) & 0xff);
         pr_buff[23] = (unsigned char)(transportid_len & 0xff);
     }
-    if (0 != do_prout(sg_fd, PROUT_REG_MOVE_SA, 0, prout_type, pr_buff,
-                      len, 1, do_verbose)) {
+    res = sg_ll_persistent_reserve_out(sg_fd, PROUT_REG_MOVE_SA, 0,
+                                       prout_type, pr_buff, len, 1,
+                                       do_verbose);
+    if (res) {
+       if (SG_LIB_CAT_INVALID_OP == res)
+            fprintf(stderr, "Persistent reserve out command not supported\n");
+        else if (SG_LIB_CAT_ILLEGAL_REQ == res)
+            fprintf(stderr, "Persistent reserve out, bad field in cdb\n");
+        else if (SG_LIB_CAT_UNIT_ATTENTION == res)
+            fprintf(stderr, "Persistent reserve out, unit attention\n");
+        else
+            fprintf(stderr, "Persistent reserve out command failed\n");
         return 1;
     } else if (do_verbose)
-        fprintf(stderr, "Persistent Reservation Out 'register and move' "
+        fprintf(stderr, "Persistent Reserve Out 'register and move' "
                 "command successful\n");
     return 0;
 }
@@ -807,13 +672,13 @@ static int build_transportid(const char * inp, unsigned char * tid_arr,
 
 int main(int argc, char * argv[])
 {
-    int sg_fd, c;
+    int sg_fd, c, res;
     unsigned int prout_type;
     unsigned long long param_rk = 0;
     unsigned long long param_sark = 0;
     unsigned int param_rtp = 0;
     char device_name[256];
-    char ebuff[EBUFF_SZ];
+    char buff[48];
     int num_prin_sa = 0;
     int num_prout_sa = 0;
     int num_prout_param = 0;
@@ -886,7 +751,7 @@ int main(int argc, char * argv[])
         case 'K':
             if (1 != sscanf(optarg, "%llx", &param_rk)) {
                 fprintf(stderr, "bad argument to '--param-rk'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             ++num_prout_param;
             break;
@@ -911,12 +776,12 @@ int main(int argc, char * argv[])
         case 'Q':
             if (1 != sscanf(optarg, "%x", &param_rtp)) {
                 fprintf(stderr, "bad argument to '--relative-target-port'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             if (param_rtp > 0xffff) {
                 fprintf(stderr, "argument to '--relative-target-port' 0 to "
                         "ffff inclusive\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             ++num_prout_param;
             break;
@@ -935,14 +800,14 @@ int main(int argc, char * argv[])
         case 'S':
             if (1 != sscanf(optarg, "%llx", &param_sark)) {
                 fprintf(stderr, "bad argument to '--param-sark'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             ++num_prout_param;
             break;
         case 'T':
             if (1 != sscanf(optarg, "%x", &prout_type)) {
                 fprintf(stderr, "bad argument to '--prout-type'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             ++num_prout_param;
             break;
@@ -962,7 +827,7 @@ int main(int argc, char * argv[])
                                        &num_transportids,
                                        sizeof(transportid_arr))) {
                 fprintf(stderr, "bad argument to '--transport-id'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             ++num_prout_param;
             break;
@@ -976,12 +841,12 @@ int main(int argc, char * argv[])
             break;
         case '?':
             usage();
-            return 1;
+            return 0;
         default:
             fprintf(stderr, "unrecognised switch "
                                 "code 0x%x ??\n", c);
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (optind < argc) {
@@ -995,37 +860,37 @@ int main(int argc, char * argv[])
                 fprintf(stderr, "Unexpected extra argument: %s\n",
                         argv[optind]);
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
 
     if ('\0' == device_name[0]) {
         fprintf(stderr, "No device name given\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     if ((want_prout + want_prin) > 1) {
         fprintf(stderr, "choose '--in' _or_ '--out' (not both)\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     } else if (want_prout) { /* syntax check on PROUT arguments */
         prin = 0;
         if ((1 != num_prout_sa) || (0 != num_prin_sa)) {
-            fprintf(stderr, ">> For Persistent Reservation Out one and "
+            fprintf(stderr, ">> For Persistent Reserve Out one and "
                     "only one appropriate\n>> service action must be "
                     "chosen (e.g. '--register')\n");
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     } else { /* syntax check on PRIN arguments */
         if (num_prout_sa > 0) {
             fprintf(stderr, ">> When a service action for Persistent "
-                    "Reservation Out is chosen the\n"
+                    "Reserve Out is chosen the\n"
                     ">> '--out' option must be given (as a safeguard)\n");
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
         if (0 == num_prin_sa) {
             fprintf(stderr, ">> No service action given; assume Persistent"
-                    " Reservations In command\n"
+                    " Reserve In command\n"
                     ">> with Read Keys service action\n");
             prin_sa = 0;
             ++num_prin_sa;
@@ -1033,20 +898,20 @@ int main(int argc, char * argv[])
             fprintf(stderr, "Too many service actions given; choose "
                     "one only\n");
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     if ((param_unreg || param_rtp) && (PROUT_REG_MOVE_SA != prout_sa)) {
         fprintf(stderr, "--unreg or --relative-target-port"
                 " only useful with --register-move\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     if ((PROUT_REG_MOVE_SA == prout_sa) && (1 != num_transportids)) {
         fprintf(stderr, "with --register-move one (and only one) "
                 "--transport-id should be given\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     if (((PROUT_RES_SA == prout_sa) ||
          (PROUT_REL_SA == prout_sa) ||
@@ -1064,17 +929,17 @@ int main(int argc, char * argv[])
     }
 
     if (do_inquiry) {
-        if ((sg_fd = open(device_name, O_RDONLY | O_NONBLOCK)) < 0) {
-            snprintf(ebuff, EBUFF_SZ, "sg_persist: error opening file: %s "
-                     " (ro)", device_name);
-            perror(ebuff);
-            return 1;
+        if ((sg_fd = sg_cmds_open_device(device_name, 1 /* ro */,
+                                         do_verbose)) < 0) {
+            fprintf(stderr, "sg_persist: error opening file (ro): %s: %s\n",
+                     device_name, safe_strerror(-sg_fd));
+            return SG_LIB_FILE_ERROR;
         }
         if (0 == sg_simple_inquiry(sg_fd, &inq_resp, 1, do_verbose)) {
             printf("  %.8s  %.16s  %.4s\n", inq_resp.vendor, inq_resp.product,
                    inq_resp.revision);
             peri_type = inq_resp.peripheral_type;
-            cp = get_ptype_str(peri_type);
+            cp = sg_get_pdt_str(peri_type, sizeof(buff), buff);
             if (strlen(cp) > 0)
                 printf("  Peripheral device type: %s\n", cp);
             else
@@ -1082,16 +947,16 @@ int main(int argc, char * argv[])
         } else {
             printf("sg_persist: %s doesn't respond to a SCSI INQUIRY\n", 
                    device_name);
-            return 1;
+            return SG_LIB_CAT_OTHER;
         }
-        close(sg_fd);
+        sg_cmds_close_device(sg_fd);
     }
 
-    if ((sg_fd = open(device_name, O_RDWR | O_NONBLOCK)) < 0) {
-        snprintf(ebuff, EBUFF_SZ, "sg_persist: error opening file: %s (rw)",
-                 device_name);
-        perror(ebuff);
-        return 1;
+    if ((sg_fd = sg_cmds_open_device(device_name, 0 /* rw */,
+                                     do_verbose)) < 0) {
+        fprintf(stderr, "sg_persist: error opening file (rw): %s: %s\n",
+                device_name, safe_strerror(-sg_fd));
+        return SG_LIB_FILE_ERROR;
     }
 
     if (prin)
@@ -1107,6 +972,11 @@ int main(int argc, char * argv[])
                          param_sark, param_alltgpt, param_aptpl,
                          transportid_arr, transportid_arr_len, do_verbose);
 
-    close(sg_fd);
-    return ret;
+    res = sg_cmds_close_device(sg_fd);
+    if (res < 0) {
+        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        if (0 == ret)
+            return SG_LIB_FILE_ERROR;
+    }
+    return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
