@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2005 Douglas Gilbert.
+ * Copyright (c) 2004-2006 Douglas Gilbert.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,12 +32,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <errno.h>
 #include <getopt.h>
-#include <sys/ioctl.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include "sg_include.h"
+
 #include "sg_lib.h"
 #include "sg_cmds.h"
 
@@ -47,7 +43,7 @@
  *commands tailored for SES (enclosure) devices.
  */
 
-static char * version_str = "1.18 20050808";
+static char * version_str = "1.27 20060623";    /* ses2r15 */
 
 #define MX_ALLOC_LEN 4096
 #define MX_ELEM_HDR 512
@@ -109,8 +105,8 @@ static void usage()
 }
 
 /* Return of 0 -> success, SG_LIB_CAT_INVALID_OP -> Send diagnostic not
- * supported, SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, -1 -> other
- * failure */
+ * supported, SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, 
+ * SG_LIB_CAT_NOT_READY, SG_LIB_CAT_UNIT_ATTENTION, -1 -> other failures */
 static int do_senddiag(int sg_fd, int pf_bit, void * outgoing_pg, 
                        int outgoing_len, int noisy, int verbose)
 {
@@ -120,39 +116,6 @@ static int do_senddiag(int sg_fd, int pf_bit, void * outgoing_pg,
                            noisy, verbose);
 }
 
-static const char * scsi_ptype_strs[] = {
-    /* 0 */ "disk",
-    "tape",
-    "printer",
-    "processor",
-    "write once optical disk",
-    /* 5 */ "cd/dvd",
-    "scanner",
-    "optical memory device",
-    "medium changer",
-    "communications",
-    /* 0xa */ "graphics",
-    "graphics",
-    "storage array controller",
-    "enclosure services device",
-    "simplified direct access device",
-    "optical card reader/writer device",
-    /* 0x10 */ "bridge controller commands",
-    "object based storage",
-    "automation/driver interface",
-    "0x13", "0x14", "0x15", "0x16", "0x17", "0x18",
-    "0x19", "0x1a", "0x1b", "0x1c", "0x1d",
-    "well known logical unit",
-    "no physical device on this lu",
-};
-
-static const char * get_ptype_str(int scsi_ptype)
-{
-    int num = sizeof(scsi_ptype_strs) / sizeof(scsi_ptype_strs[0]);
-
-    return (scsi_ptype < num) ? scsi_ptype_strs[scsi_ptype] : "";
-}
-
 struct page_code_desc {
         int page_code;
         const char * desc;
@@ -178,6 +141,27 @@ static struct page_code_desc pc_desc_arr[] = {
         {0x40, "Translate address (SBC)"},
         {0x41, "Device status (SBC)"},
 };
+static struct page_code_desc in_pc_desc_arr[] = {
+        {0x0, "Supported diagnostic pages"},
+        {0x1, "Configuration (SES)"},
+        {0x2, "Enclosure status (SES)"},
+        {0x3, "Help text (SES)"},
+        {0x4, "String In (SES)"},
+        {0x5, "Threshold In (SES)"},
+        {0x6, "Array Status (SES, obsolete)"},
+        {0x7, "Element descriptor (SES)"},
+        {0x8, "Short enclosure status (SES)"},
+        {0x9, "Enclosure busy (SES-2)"},
+        {0xa, "Additional (device) element status (SES-2)"},
+        {0xb, "Subenclosure help text (SES-2)"},
+        {0xc, "Subenclosure string In (SES-2)"},
+        {0xd, "Supported SES diagnostic pages (SES-2)"},
+        {0xe, "Download microcode (SES-2)"},
+        {0xf, "Subenclosure nickname (SES-2)"},
+        {0x3f, "Protocol specific SAS (SAS-1)"},
+        {0x40, "Translate address (SBC)"},
+        {0x41, "Device status (SBC)"},
+};
 
 static const char * find_page_code_desc(int page_num)
 {
@@ -194,6 +178,21 @@ static const char * find_page_code_desc(int page_num)
     return NULL;
 }
 
+static const char * find_in_page_code_desc(int page_num)
+{
+    int k;
+    int num = sizeof(in_pc_desc_arr) / sizeof(in_pc_desc_arr[0]);
+    const struct page_code_desc * pcdp = &in_pc_desc_arr[0];
+
+    for (k = 0; k < num; ++k, ++pcdp) {
+        if (page_num == pcdp->page_code)
+            return pcdp->desc;
+        else if (page_num < pcdp->page_code)
+            return NULL;
+    }
+    return NULL;
+}
+
 struct element_desc {
         int type_code;
         const char * desc;
@@ -285,9 +284,9 @@ static void ses_configuration_sdg(const unsigned char * resp, int resp_len)
             fprintf(stderr, "      enc descriptor len=%d ??\n", el);
             continue;
         }
-        printf("      logical id (hex): ");
+        printf("      logical identifier (hex): ");
         for (j = 0; j < 8; ++j)
-            printf("%02x ", ucp[4 + j]);
+            printf("%02x", ucp[4 + j]);
         printf("\n      vendor: %.8s  product: %.16s  rev: %.4s\n",
                ucp + 12, ucp + 20, ucp + 36);
         if (el > 40) {
@@ -321,20 +320,23 @@ truncated:
     return;
 }
 
+/* Returns number of elements written to 'ehp' or -1 if there is
+   a problem */ 
 static int populate_element_hdr_arr(int fd, struct element_hdr * ehp,
                                     unsigned int * generationp, int verbose)
 {
-    int resp_len, k, el, num_subs, sum_elem_types;
+    int resp_len, k, el, num_subs, sum_elem_types, res;
     unsigned int gen_code;
     unsigned char resp[MX_ALLOC_LEN];
     int rsp_buff_size = MX_ALLOC_LEN;
     const unsigned char * ucp;
     const unsigned char * last_ucp;
 
-    if (0 == sg_ll_receive_diag(fd, 1, 1, resp, rsp_buff_size, 1, verbose)) {
+    res = sg_ll_receive_diag(fd, 1, 1, resp, rsp_buff_size, 1, verbose);
+    if (0 == res) {
         resp_len = (resp[2] << 8) + resp[3] + 4;
         if (resp_len > rsp_buff_size) {
-            fprintf(stderr, "<<< warning response buffer too small "
+            fprintf(stderr, "<<< warning: response buffer too small "
                     "[%d but need %d]>>>\n", rsp_buff_size, resp_len);
             resp_len = rsp_buff_size;
         }
@@ -384,7 +386,7 @@ static int populate_element_hdr_arr(int fd, struct element_hdr * ehp,
         }
         return sum_elem_types;
     } else {
-        fprintf(stderr, "populate: couldn't read config page\n");
+        fprintf(stderr, "populate: couldn't read config page, res=%d\n", res);
         return -1;
     }
 p_truncated:
@@ -400,34 +402,39 @@ static char * find_sas_connector_type(int conn_type, char * buff,
         snprintf(buff, buff_len, "No information");
         break;
     case 0x1:
-        snprintf(buff, buff_len, "SAS external receptacle (SFF-8470) "
+        snprintf(buff, buff_len, "SAS 4x receptacle (SFF-8470) "
                  "[max 4 phys]");
         break;
     case 0x2:
-        snprintf(buff, buff_len, "SAS external compact receptacle "
-                 "(SFF-8088) [max 4 phys]");
-        break;
-    case 0x10:
-        snprintf(buff, buff_len, "SAS internal wide plug (SFF-8484) "
+        snprintf(buff, buff_len, "Mini SAS 4x receptacle (SFF-8088) "
                  "[max 4 phys]");
         break;
+    case 0xf:
+        snprintf(buff, buff_len, "Vendor specific external connector");
+        break;
+    case 0x10:
+        snprintf(buff, buff_len, "SAS 4i plug (SFF-8484) [max 4 phys]");
+        break;
     case 0x11:
-        snprintf(buff, buff_len, "SAS internal compact wide plug "
-                 "(SFF-8087) [max 4 phys]");
+        snprintf(buff, buff_len, "Mini SAS 4i receptacle (SFF-8087) "
+                 "[max 4 phys]");
         break;
     case 0x20:
-        snprintf(buff, buff_len, "SAS backplane receptacle (SFF-8482) "
+        snprintf(buff, buff_len, "SAS Drive backplane receptacle (SFF-8482) "
                  "[max 2 phys]");
         break;
     case 0x21:
-        snprintf(buff, buff_len, "SATA style host plug [max 1 phy]");
+        snprintf(buff, buff_len, "SATA host plug [max 1 phy]");
         break;
     case 0x22:
-        snprintf(buff, buff_len, "SAS plug (SFF-8482) [max 2 phys]");
+        snprintf(buff, buff_len, "SAS Drive plug (SFF-8482) [max 2 phys]");
         break;
     case 0x23:
         snprintf(buff, buff_len, "SATA device plug [max 1 phy]");
         break;
+    case 0x3f:
+        snprintf(buff, buff_len, "Vendor specific internal connector");
+        break;
     default:
         if (conn_type < 0x10)
             snprintf(buff, buff_len, "unknown external connector type: 0x%x",
@@ -444,7 +451,7 @@ static char * find_sas_connector_type(int conn_type, char * buff,
         else if (conn_type < 0x80)
             snprintf(buff, buff_len, "vendor specific connector type: 0x%x",
                      conn_type);
-        else    /* conn_type comes from a 7 bit field, so this is imposs */
+        else    /* conn_type is a 7 bit field, so this is impossible */
             snprintf(buff, buff_len, "unexpected connector type: 0x%x",
                      conn_type);
         break;
@@ -531,7 +538,7 @@ static void print_element_status(const char * pad,
                    !!(statp[1] & 0x80), !!(statp[3] & 0x40),
                    !!(statp[3] & 0x20), !!(statp[3] & 0x10));
         printf("%sActual speed=%d rpm, Fan %s\n", pad,
-               (((3 & statp[1]) << 8) + statp[2]) * 10,
+               (((0x7 & statp[1]) << 8) + statp[2]) * 10,
                actual_speed_desc[7 & statp[3]]);
         break;
     case 4:     /* temperature sensor */
@@ -613,13 +620,16 @@ static void print_element_status(const char * pad,
                    !!(statp[2] & 0x8));
         if ((! filter) || ((0x7 & statp[2]) || (0x83 & statp[3])))
             printf("%sUPS fail=%d, Warn=%d, Intf fail=%d, Ident=%d, Batt fail"
-                   "=%d,BPF=%d\n", pad, !!(statp[2] & 0x4), !!(statp[2] & 0x2),
+                   "=%d, BPF=%d\n", pad, !!(statp[2] & 0x4), !!(statp[2] & 0x2),
                    !!(statp[2] & 0x1), !!(statp[3] & 0x80), !!(statp[3] & 0x2),
                    !!(statp[3] & 0x1));
         break;
-    case 0xc:   /* Display */
+    case 0xc:   /* Display (ses2r15) */
         if ((! filter) || (0x80 & statp[1]))
-            printf("%sIdent=%d\n", pad, !!(statp[1] & 0x80));
+            printf("%sIdent=%d, Display mode status=%d, Display "
+                   "character status=0x%x\n", pad,
+                   !!(statp[1] & 0x80), (statp[1] & 0x3),
+                   ((statp[2] << 8) & statp[3]));
         break;
     case 0xd:   /* Key pad entry */
         if ((! filter) || (0x80 & statp[1]))
@@ -993,23 +1003,23 @@ static char * sas_device_type[] = {
     "reserved [4]", "reserved [5]", "reserved [6]", "reserved [7]"
 };
 
-static void ses_transport_proto(const unsigned char * ucp, int len,
-                                int elem_num)
+static void ses_additional_elem_each(const unsigned char * ucp, int len,
+                                     int elem_num, int elem_type)
 {
-    int ports, phys, j, m, desc_type, eip_off;
+    int ports, phys, j, m, desc_type, eip_offset;
     const unsigned char * per_ucp;
 
-    eip_off = (0x10 & ucp[0]) ? 2 : 0;
+    eip_offset = (0x10 & ucp[0]) ? 2 : 0;
     switch (0xf & ucp[0]) {
     case 0:     /* FCP */
-        ports = ucp[2 + eip_off];
+        ports = ucp[2 + eip_offset];
         printf("   [%d] Transport protocol: FCP, number of ports: %d\n",
                elem_num + 1, ports);
         printf("    node_name: ");
         for (m = 0; m < 8; ++m)
-            printf("%02x", ucp[6 + eip_off + m]);
+            printf("%02x", ucp[6 + eip_offset + m]);
         printf("\n");
-        per_ucp = ucp + 14 + eip_off;
+        per_ucp = ucp + 14 + eip_offset;
         for (j = 0; j < ports; ++j, per_ucp += 16) {
             printf("    [%d] port loop position: %d, port requested hard "
                    "address: %d\n", j + 1, per_ucp[0], per_ucp[4]);
@@ -1022,14 +1032,14 @@ static void ses_transport_proto(const unsigned char * ucp, int len,
         }
         break;
     case 6:     /* SAS */
-        desc_type = (ucp[3 + eip_off] >> 6) & 0x3;
+        desc_type = (ucp[3 + eip_offset] >> 6) & 0x3;
         printf("   [%d] Transport protocol: SAS, ", elem_num + 1);
         if (0 == desc_type) {
-            phys = ucp[2 + eip_off];
+            phys = ucp[2 + eip_offset];
             printf("SAS and SATA device descriptor type [%d]\n", desc_type);
             printf("    number of phys: %d, not all phys: %d\n", phys,
-                   ucp[3 + eip_off] & 1);
-            per_ucp = ucp + 4 + eip_off + eip_off;
+                   ucp[3 + eip_offset] & 1);
+            per_ucp = ucp + 4 + eip_offset + eip_offset;
             for (j = 0; j < phys; ++j, per_ucp += 28) {
                 printf("    [%d] device type: %s\n", phys + 1,
                        sas_device_type[(0x70 & per_ucp[0]) >> 4]);
@@ -1043,33 +1053,57 @@ static void ses_transport_proto(const unsigned char * ucp, int len,
                        ((per_ucp[3] & 4) ? "STP" : ""),
                        ((per_ucp[3] & 2) ? "SMP" : ""),
                        ((per_ucp[3] & 1) ? "SATA_device" : ""));
-                printf("      attached SAS address: ");
+                printf("      attached SAS address: 0x");
                 for (m = 0; m < 8; ++m)
                     printf("%02x", per_ucp[4 + m]);
-                printf("\n      SAS address: ");
+                printf("\n      SAS address: 0x");
                 for (m = 0; m < 8; ++m)
                     printf("%02x", per_ucp[12 + m]);
                 printf("\n      phy identifier: 0x%x\n", per_ucp[20]);
             }
         } else if (1 == desc_type) {
-            phys = ucp[2 + eip_off];
-            printf("expander descriptor type [%d]\n", desc_type);
-            printf("    number of phys: %d\n", phys);
-            printf("    SAS address: ");
-            for (m = 0; m < 8; ++m)
-                printf("%02x", ucp[6 + eip_off + m]);
-            printf("\n");
-            per_ucp = ucp + 14 + eip_off;
-            for (j = 0; j < phys; ++j, per_ucp += 2) {
-                printf("    [%d] ", phys + 1);
-                if (0xff == per_ucp[0])
-                    printf("no attached connector");
-                else
-                    printf("connector element index: %d", per_ucp[0]);
-                if (0xff != per_ucp[1])
-                    printf(", other element index: %d", per_ucp[1]);
+            phys = ucp[2 + eip_offset];
+            if (0x18 == elem_type) {
+                printf("expander descriptor type\n");
+                printf("    number of phys: %d\n", phys);
+                printf("    SAS address: 0x");
+                for (m = 0; m < 8; ++m)
+                    printf("%02x", ucp[6 + eip_offset + m]);
                 printf("\n");
-            }
+                per_ucp = ucp + 14 + eip_offset;
+                for (j = 0; j < phys; ++j, per_ucp += 2) {
+                    printf("      [%d] ", j + 1);
+                    if (0xff == per_ucp[0])
+                        printf("no attached connector");
+                    else
+                        printf("connector element index: %d", per_ucp[0]);
+                    if (0xff != per_ucp[1])
+                        printf(", other element index: %d", per_ucp[1]);
+                    printf("\n");
+                }
+            } else if ((0x14 == elem_type) || (0x15 == elem_type)) {
+                printf("SCSI %s port descriptor type\n", 
+                       ((0x14 == elem_type) ? "target" : "initiator"));
+                printf("    number of phys: %d\n", phys);
+                per_ucp = ucp + 6 + eip_offset;
+                for (j = 0; j < phys; ++j, per_ucp += 12) {
+                    printf("    phy identifier: 0x%x\n", per_ucp[0]);
+                    if (0xff == per_ucp[2])
+                        printf("      no attached connector");
+                    else
+                        printf("      connector element index: %d",
+                               per_ucp[2]);
+                    if (0xff != per_ucp[3])
+                        printf(", other element index: %d", per_ucp[3]);
+                    printf("\n");
+                    printf("      SAS address: 0x");
+                    for (m = 0; m < 8; ++m)
+                        printf("%02x", per_ucp[4 + m]);
+                    printf("\n");
+                }
+            } else
+                printf("    unrecognised element type [%d] for desc_type 1\n",
+                       elem_type);
         } else
             printf("    unrecognised descriptor type [%d]\n", desc_type);
         break;
@@ -1082,7 +1116,7 @@ static void ses_transport_proto(const unsigned char * ucp, int len,
 }
 
 /* Previously called "Device element status descriptor". Changed "device"
-   to "additional" to allow for SAS descriptors and SATA devices */
+   to "additional" to allow for SAS expander and SATA devices */
 static void ses_additional_elem_sdg(const struct element_hdr * ehp,
                          int num_telems, unsigned int ref_gen_code,
                          const unsigned char * resp, int resp_len)
@@ -1110,9 +1144,12 @@ static void ses_additional_elem_sdg(const struct element_hdr * ehp,
         if ((ucp + 1) > last_ucp)
             goto truncated;
         elem_type = ehp[k].etype;
-        if (! ((1 == elem_type) || (0x17 == elem_type) || (0x18 != elem_type)))
-            continue;
-        /* only interested in device, array and SAS expander elements */
+        if (! ((1 == elem_type) ||      /* device */
+               (0x14 == elem_type) ||   /* scsi target */
+               (0x15 == elem_type) ||   /* scsi initiator */
+               (0x17 == elem_type) ||   /* array */
+               (0x18 == elem_type)))    /* SAS expander */
+            continue;   /* skip if not one of above element types */
         cp = find_element_desc(elem_type);
         if (cp)
             printf("  Element type: %s, subenclosure id: %d\n",
@@ -1127,10 +1164,10 @@ static void ses_additional_elem_sdg(const struct element_hdr * ehp,
             if (0x6 == proto) {  /* SAS */
                 desc_type = ((eip ? ucp[5] : ucp[3]) >> 6) & 0x3;
                 if (eip) {
-                    if (0 == desc_type) /* non expander */
+                    if (0 == desc_type) /* device or array element */
                         printf("    element index: %d [0x%x], bay number: %d"
                                " [0x%x]\n", ucp[3], ucp[3], ucp[7], ucp[7]);
-                    else
+                    else  /* target, initiator or SAS expander element */
                         printf("    element index: %d [0x%x]\n", ucp[3],
                                ucp[3]);
                 }
@@ -1142,7 +1179,7 @@ static void ses_additional_elem_sdg(const struct element_hdr * ehp,
             if (invalid)
                 printf("      flagged as invalid (no further information)\n");
             else
-                ses_transport_proto(ucp, desc_len, j);
+                ses_additional_elem_each(ucp, desc_len, j, elem_type);
         }
     }
     return;
@@ -1369,9 +1406,9 @@ static int read_hex(const char * inp, unsigned char * arr, int * arr_len)
     return 0;
 }
 
-static void ses_process_status(int sg_fd, int page_code, int do_raw,
-                               int do_hex, int inner_hex, int filter,
-                               int verbose)
+static int ses_process_status(int sg_fd, int page_code, int do_raw,
+                              int do_hex, int inner_hex, int filter,
+                              int verbose)
 {
     int rsp_len, res;
     unsigned int ref_gen_code;
@@ -1380,9 +1417,10 @@ static void ses_process_status(int sg_fd, int page_code, int do_raw,
     const char * cp;
 
     memset(rsp_buff, 0, rsp_buff_size);
-    cp = find_page_code_desc(page_code);
-    if (0 == sg_ll_receive_diag(sg_fd, 1, page_code, rsp_buff,
-                                rsp_buff_size, 1, verbose)) {
+    cp = find_in_page_code_desc(page_code);
+    res = sg_ll_receive_diag(sg_fd, 1, page_code, rsp_buff,
+                             rsp_buff_size, 1, verbose);
+    if (0 == res) {
         rsp_len = (rsp_buff[2] << 8) + rsp_buff[3] + 4;
         if (rsp_len > rsp_buff_size) {
             fprintf(stderr, "<<< warning response buffer too small "
@@ -1506,6 +1544,7 @@ static void ses_process_status(int sg_fd, int page_code, int do_raw,
             printf("Attempt to fetch status diagnostic page [0x%x] "
                    "failed\n", page_code);
     }
+    return res;
 }
 
 
@@ -1524,10 +1563,11 @@ int main(int argc, char * argv[])
     int inner_hex = 0;
     int byte1 = 0;
     char device_name[256];
+    char buff[48];
     unsigned char data_arr[1024];
     int arr_len = 0;
     int pd_type = 0;
-    int ret = 1;
+    int ret = 0;
     struct sg_simple_inquiry_resp inq_resp;
     const char * cp;
 
@@ -1546,7 +1586,7 @@ int main(int argc, char * argv[])
             if ((byte1 < 0) || (byte1 > 255)) {
                 fprintf(stderr, "bad argument to '--byte1' (0 to 255 "
                         "inclusive)\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 'c':
@@ -1556,7 +1596,7 @@ int main(int argc, char * argv[])
             memset(data_arr, 0, sizeof(data_arr));
             if (read_hex(optarg, data_arr + 4, &arr_len)) {
                 fprintf(stderr, "bad argument to '--data'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             do_data = 1;
             break;
@@ -1581,7 +1621,7 @@ int main(int argc, char * argv[])
             if ((page_code < 0) || (page_code > 255)) {
                 fprintf(stderr, "bad argument to '--page' (0 to 255 "
                         "inclusive)\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 'r':
@@ -1599,7 +1639,7 @@ int main(int argc, char * argv[])
         default:
             fprintf(stderr, "unrecognised switch code 0x%x ??\n", c);
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (optind < argc) {
@@ -1613,7 +1653,7 @@ int main(int argc, char * argv[])
                 fprintf(stderr, "Unexpected extra argument: %s\n",
                         argv[optind]);
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (do_list) {
@@ -1636,12 +1676,12 @@ int main(int argc, char * argv[])
     if (do_control && do_status) {
         fprintf(stderr, "cannot have both '--control' and '--status'\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     } else if (1 == do_control) {
         if (! do_data) {
             fprintf(stderr, "need to give '--data' in control mode\n");
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     } else if (0 == do_status)
         do_status = 1;  /* default to receiving status pages */
@@ -1649,25 +1689,26 @@ int main(int argc, char * argv[])
     if (0 == device_name[0]) {
         fprintf(stderr, "missing device name!\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
 
-    sg_fd = open(device_name, O_RDWR | O_NONBLOCK);
+    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
     if (sg_fd < 0) {
-        fprintf(stderr, ME "open error: %s: ", device_name);
-        perror("");
-        return 1;
+        fprintf(stderr, ME "open error: %s: %s\n", device_name,
+                safe_strerror(-sg_fd));
+        return SG_LIB_FILE_ERROR;
     }
     if (! do_raw) {
         if (sg_simple_inquiry(sg_fd, &inq_resp, 1, verbose)) {
             fprintf(stderr, ME "%s doesn't respond to a SCSI INQUIRY\n",
                     device_name);
+            ret = SG_LIB_CAT_OTHER;
             goto err_out;
         } else {
             printf("  %.8s  %.16s  %.4s\n", inq_resp.vendor,
                    inq_resp.product, inq_resp.revision);
             pd_type = inq_resp.peripheral_type;
-            cp = get_ptype_str(pd_type);
+            cp = sg_get_pdt_str(pd_type, sizeof(buff), buff);
             if (0xd == pd_type)
                 printf("    enclosure services device\n");
             else if (0x40 & inq_resp.byte_6)
@@ -1676,10 +1717,10 @@ int main(int argc, char * argv[])
                 printf("    %s device (not an enclosure)\n", cp);
         }
     }
-    if (do_status)
-        ses_process_status(sg_fd, page_code, do_raw, do_hex, inner_hex,
-                           do_filter, verbose);
-    else { /* control page requested */
+    if (do_status) {
+        ret = ses_process_status(sg_fd, page_code, do_raw, do_hex,
+                                 inner_hex, do_filter, verbose);
+    } else { /* control page requested */
         data_arr[0] = page_code;
         data_arr[1] = byte1;
         data_arr[2] = (arr_len >> 8) & 0xff;
@@ -1688,7 +1729,8 @@ int main(int argc, char * argv[])
         case 0x2:       /* Enclosure control diagnostic page */
             printf("Sending Enclosure control [0x%x] page, with page "
                    "length=%d bytes\n", page_code, arr_len);
-            if (do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose)) {
+            ret = do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose);
+            if (ret) {
                 fprintf(stderr, "couldn't send Enclosure control page\n");
                 goto err_out;
             }
@@ -1696,7 +1738,8 @@ int main(int argc, char * argv[])
         case 0x4:       /* String Out diagnostic page */
             printf("Sending String Out [0x%x] page, with page length=%d "
                    "bytes\n", page_code, arr_len);
-            if (do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose)) {
+            ret = do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose);
+            if (ret) {
                 fprintf(stderr, "couldn't send String Out page\n");
                 goto err_out;
             }
@@ -1704,7 +1747,8 @@ int main(int argc, char * argv[])
         case 0x5:       /* Threshold Out diagnostic page */
             printf("Sending Threshold Out [0x%x] page, with page length=%d "
                    "bytes\n", page_code, arr_len);
-            if (do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose)) {
+            ret = do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose);
+            if (ret) {
                 fprintf(stderr, "couldn't send Threshold Out page\n");
                 goto err_out;
             }
@@ -1712,7 +1756,8 @@ int main(int argc, char * argv[])
         case 0x6:       /* Array control diagnostic page (obsolete) */
             printf("Sending Array control [0x%x] page, with page "
                    "length=%d bytes\n", page_code, arr_len);
-            if (do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose)) {
+            ret = do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose);
+            if (ret) {
                 fprintf(stderr, "couldn't send Array control page\n");
                 goto err_out;
             }
@@ -1720,7 +1765,8 @@ int main(int argc, char * argv[])
         case 0xc:       /* Subenclosure String Out diagnostic page */
             printf("Sending Subenclosure String Out [0x%x] page, with page "
                    "length=%d bytes\n", page_code, arr_len);
-            if (do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose)) {
+            ret = do_senddiag(sg_fd, 1, data_arr, arr_len + 4, 1, verbose);
+            if (ret) {
                 fprintf(stderr, "couldn't send Subenclosure String Out "
                         "page\n");
                 goto err_out;
@@ -1729,16 +1775,17 @@ int main(int argc, char * argv[])
         default:
             fprintf(stderr, "Setting SES control page 0x%x not supported "
                     "yet\n", page_code);
+            ret = SG_LIB_SYNTAX_ERROR;
             break;
         }
     }
 
-    ret = 0;
 err_out:
-    res = close(sg_fd);
+    res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
-        perror(ME "close error");
-        return 2;
+        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        if (0 == ret)
+            return SG_LIB_FILE_ERROR;
     }
-    return ret;
+    return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
