@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2005 Douglas Gilbert.
+ * Copyright (c) 2004-2006 Douglas Gilbert.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,11 +32,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <errno.h>
 #include <getopt.h>
-#include <sys/ioctl.h>
-#include <sys/types.h>
-#include "sg_include.h"
+
 #include "sg_lib.h"
 #include "sg_cmds.h"
 
@@ -46,7 +43,7 @@
  * mode page on the given device.
  */
 
-static char * version_str = "1.04 20050722";
+static char * version_str = "1.06 20060623";
 
 #define ME "sg_wr_mode: "
 
@@ -280,7 +277,7 @@ static int build_mask(const char * inp, unsigned char * mask_arr,
 
 int main(int argc, char * argv[])
 {
-    int sg_fd, res, c, num, alloc_len, off;
+    int sg_fd, res, c, num, alloc_len, off, pdt;
     int k, md_len, hdr_len, bd_len, mask_in_len;
     unsigned u, uu;
     int dbd = 0;
@@ -298,7 +295,8 @@ int main(int argc, char * argv[])
     unsigned char mask_in[MX_ALLOC_LEN];
     unsigned char ref_md[MX_ALLOC_LEN];
     char ebuff[EBUFF_SZ];
-    int ret = 1;
+    struct sg_simple_inquiry_resp inq_data;
+    int ret = 0;
 
     memset(device_name, 0, sizeof device_name);
     while (1) {
@@ -315,7 +313,7 @@ int main(int argc, char * argv[])
             if (0 != build_mode_page(optarg, read_in, &read_in_len,
                                      sizeof(read_in))) {
                 fprintf(stderr, "bad argument to '--contents'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             got_contents = 1;
             break;
@@ -335,7 +333,7 @@ int main(int argc, char * argv[])
                 mode_6 = (6 == res) ? 1 : 0;
             else {
                 fprintf(stderr, "length (of cdb) must be 6 or 10\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 'm':
@@ -343,7 +341,7 @@ int main(int argc, char * argv[])
             if (0 != build_mask(optarg, mask_in, &mask_in_len,
                                 sizeof(mask_in))) {
                 fprintf(stderr, "bad argument to '--mask'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             got_mask = 1;
             break;
@@ -353,21 +351,21 @@ int main(int argc, char * argv[])
                 if ((1 != num) || (u > 62)) {
                     fprintf(stderr, "Bad page code value after '--page' "
                             "switch\n");
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
                 pg_code = u;
             } else if (2 == sscanf(optarg, "%x,%x", &u, &uu)) {
                 if (uu > 254) {
                     fprintf(stderr, "Bad sub page code value after '--page'"
                             " switch\n");
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
                 pg_code = u;
                 sub_pg_code = uu;
             } else {
                 fprintf(stderr, "Bad page code, subpage code sequence after "
                         "'--page' switch\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 's':
@@ -382,7 +380,7 @@ int main(int argc, char * argv[])
         default:
             fprintf(stderr, "unrecognised switch code 0x%x ??\n", c);
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (optind < argc) {
@@ -396,31 +394,35 @@ int main(int argc, char * argv[])
                 fprintf(stderr, "Unexpected extra argument: %s\n",
                         argv[optind]);
             usage();
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (0 == device_name[0]) {
         fprintf(stderr, "missing device name!\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     if (pg_code < 0) {
         fprintf(stderr, "need page code (see '--page=')\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     if (got_mask && force) {
         fprintf(stderr, "cannot use both '--force' and '--mask'\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
 
-    sg_fd = open(device_name, O_RDWR | O_NONBLOCK);
+    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
     if (sg_fd < 0) {
-        fprintf(stderr, ME "open error: %s: ", device_name);
-        perror("");
-        return 1;
+        fprintf(stderr, ME "open error: %s: %s\n", device_name,
+                safe_strerror(-sg_fd));
+        return SG_LIB_FILE_ERROR;
     }
+    if (0 == sg_simple_inquiry(sg_fd, &inq_data, 0, verbose))
+        pdt = inq_data.peripheral_type;
+    else
+        pdt = 0x1f;
 
     /* do MODE SENSE to fetch current values */
     memset(ref_md, 0, MX_ALLOC_LEN);
@@ -432,10 +434,19 @@ int main(int argc, char * argv[])
         res = sg_ll_mode_sense10(sg_fd, 0 /* llbaa */, dbd, 0 /* current */,
                                  pg_code, sub_pg_code, ref_md, alloc_len, 1,
                                  verbose);
+    ret = res;
     if (SG_LIB_CAT_INVALID_OP == res) {
         fprintf(stderr, "MODE SENSE (%d) not supported, try '--len=%d'\n",
                 (mode_6 ? 6 : 10), (mode_6 ? 10 : 6));
         goto err_out;
+    } else if (SG_LIB_CAT_NOT_READY == res) {
+        fprintf(stderr, "MODE SENSE (%d) failed, device not ready\n",
+                (mode_6 ? 6 : 10));
+        goto err_out;
+    } else if (SG_LIB_CAT_UNIT_ATTENTION == res) {
+        fprintf(stderr, "MODE SENSE (%d) failed, unit attention\n",
+                (mode_6 ? 6 : 10));
+        goto err_out;
     } else if (SG_LIB_CAT_ILLEGAL_REQ == res) {
         fprintf(stderr, "bad field in MODE SENSE (%d) command\n",
                 (mode_6 ? 6 : 10));
@@ -466,6 +477,8 @@ int main(int argc, char * argv[])
         ref_md[0] = 0;  /* mode data length reserved for mode select */
         if (! mode_6)
             ref_md[1] = 0;    /* mode data length reserved for mode select */
+        if (0 == pdt)   /* for disks mask out DPOFUA bit */
+            ref_md[mode_6 ? 2 : 3] &= 0xef;
         if (md_len > alloc_len) {
             fprintf(stderr, "mode data length=%d exceeds allocation "
                     "length=%d\n", md_len, alloc_len);
@@ -519,10 +532,19 @@ int main(int argc, char * argv[])
         else
             res = sg_ll_mode_select10(sg_fd, 1, save, ref_md, md_len, 1,
                                       verbose);
+        ret = res;
         if (SG_LIB_CAT_INVALID_OP == res) {
             fprintf(stderr, "MODE SELECT (%d) not supported\n",
                     (mode_6 ? 6 : 10));
             goto err_out;
+        } else if (SG_LIB_CAT_NOT_READY == res) {
+            fprintf(stderr, "MODE SELECT (%d) failed, device not ready\n",
+                    (mode_6 ? 6 : 10));
+            goto err_out;
+        } else if (SG_LIB_CAT_UNIT_ATTENTION == res) {
+            fprintf(stderr, "MODE SELECT (%d) failed, unit attention\n",
+                    (mode_6 ? 6 : 10));
+            goto err_out;
         } else if (SG_LIB_CAT_ILLEGAL_REQ == res) {
             fprintf(stderr, "bad field in MODE SELECT (%d) command\n",
                     (mode_6 ? 6 : 10));
@@ -543,12 +565,12 @@ int main(int argc, char * argv[])
         printf("  mode page:\n");
         dStrHex((const char *)(ref_md + off), md_len - off, -1);
     }
-    ret = 0;
 err_out:
-    res = close(sg_fd);
+    res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
-        perror(ME "close error");
-        return 1;
+        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        if (0 == ret)
+            return SG_LIB_FILE_ERROR;
     }
-    return ret;
+    return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
