@@ -18,15 +18,15 @@
 #include <sys/time.h>
 #include <linux/major.h> 
 #include <linux/fs.h>   /* <sys/mount.h> */
-#include "sg_include.h"
 #include "sg_lib.h"
 #include "sg_cmds.h"
+#include "sg_io_linux.h"
 #include "llseek.h"
 
 /* A utility program for copying files. Specialised for "files" that
 *  represent devices that understand the SCSI command set.
 *
-*  Copyright (C) 1999 - 2005 D. Gilbert and P. Allworth
+*  Copyright (C) 1999 - 2006 D. Gilbert and P. Allworth
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -54,10 +54,11 @@
    This version is designed for the linux kernel 2.4 and 2.6 series.
 */
 
-static char * version_str = "1.21 20050909";
+static char * version_str = "1.26 20060625";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_BLOCKS_PER_TRANSFER 128
+#define DEF_BLOCKS_PER_2048TRANSFER 32
 #define DEF_SCSI_CDBSZ 10
 #define MAX_SCSI_CDBSZ 16
 
@@ -92,9 +93,12 @@ static char * version_str = "1.21 20050909";
 #define FT_DEV_NULL 8           /* either "/dev/null" or "." as filename */
 #define FT_ST 16                /* filetype is st char device (tape) */
 #define FT_BLOCK 32             /* filetype is a block device */
+#define FT_ERROR 64             /* couldn't "stat" file */
 
 #define DEV_NULL_MINOR_NUM 3
 
+#define MIN_RESERVED_SIZE 8192
+
 static int sum_of_resids = 0;
 
 static long long dd_count = -1;
@@ -106,6 +110,15 @@ static int verbose = 0;
 
 static const char * proc_allow_dio = "/proc/scsi/sg/allow_dio";
 
+struct flags_t {
+    int append;
+    int direct;
+    int dpo;
+    int dsync;
+    int excl;
+    int fua;
+};
+
 
 static void install_handler (int sig_num, void (*sig_handler) (int sig))
 {
@@ -146,7 +159,7 @@ static void siginfo_handler(int sig)
 {
     sig = sig;  /* dummy to stop -W warning messages */
     fprintf(stderr, "Progress report, continuing ...\n");
-    print_stats ();
+    print_stats();
 }
 
 int dd_filetype(const char * filename)
@@ -157,7 +170,7 @@ int dd_filetype(const char * filename)
     if ((1 == len) && ('.' == filename[0]))
         return FT_DEV_NULL;
     if (stat(filename, &st) < 0)
-        return FT_OTHER;
+        return FT_ERROR;
     if (S_ISCHR(st.st_mode)) {
         if ((MEM_MAJOR == major(st.st_rdev)) &&
             (DEV_NULL_MINOR_NUM == minor(st.st_rdev)))
@@ -188,20 +201,24 @@ static char * dd_filetype_str(int ft, char * buff)
     if (FT_RAW & ft)
         off += snprintf(buff + off, 32, "raw device ");
     if (FT_OTHER & ft)
-        off += snprintf(buff + off, 32, "other (perhaps name file) ");
+        off += snprintf(buff + off, 32, "other (perhaps ordinary file) ");
+    if (FT_ERROR & ft)
+        off += snprintf(buff + off, 32, "unable to 'stat' file ");
     return buff;
 }
 
 void usage()
 {
-    fprintf(stderr, "Usage: "
-           "sgm_dd  [if=<infile>] [skip=<n>] [of=<ofile>] [seek=<n>] "
-           "[bs=<num>]\n"
-           "               [count=<n>] [ibs=<n>] [obs=<n>] [bpt=<num>] "
-           "[cdbsz=6|10|12|16]\n"
-           "               [dio=0|1] [fua=0|1|2|3] [sync=0|1] [time=<n>] "
-           "[verbose=<n>]\n"
-           "               [--version]\n"
+   fprintf(stderr, "Usage: "
+           "sgm_dd  [bs=<n>] [count=<n>] [ibs=<n>] [if=<ifile>]"
+           " [iflag=<flags>]\n"
+           "               [obs=<n>] [of=<ofile>] [oflag=<flags>] "
+           "[seek=<n>] [skip=<n>]\n"
+           "               [--help] [--version]\n\n");
+    fprintf(stderr,
+           "               [bpt=<num>] [cdbsz=6|10|12|16] [dio=0|1] "
+           "[fua=0|1|2|3]\n"
+           "               [sync=0|1] [time=<n>] [verbose=<n>]\n\n"
            "  bpt     is blocks_per_transfer (default is 128)\n"
            "  bs      must be device block size (default 512)\n"
            "  cdbsz   size of SCSI READ or WRITE command (default is 10)\n"
@@ -210,14 +227,10 @@ void usage()
            "  fua     force unit access: 0->don't(def), 1->of, 2->if, "
            "3->of+if\n");
     fprintf(stderr, 
-           "  ibs     input block size (if given must be same as 'bs')\n"
-           "  if      file or device to read from (def stdin)\n"
-           "  obs     output block size (if given must be same as 'bs')\n"
-           "  of      file or device to write to (def stdout), name '.' "
-           "translated to\n"
-           "          /dev/null\n"
-           "  seek    block position to start writing to 'of'\n"
-           "  skip    block position to start reading from 'if'\n"
+           "  iflag   comma separated list from: [direct,dpo,dsync,excl,"
+           "fua]\n"
+           "  oflag   comma separated list from: [append,direct,dpo,"
+           "dsync,excl,fua]\n"
            "  sync    0->no sync(def), 1->SYNCHRONIZE CACHE after xfer\n"
            "  time    0->no timing(def), 1->time plus calculate "
            "throughput\n"
@@ -225,14 +238,13 @@ void usage()
            "  --version  print version information then exit\n");
 }
 
-/* Return of 0 -> success, SG_LIB_CAT_INVALID_OP -> invalid opcode,
- * SG_LIB_CAT_MEDIA_CHANGED -> media changed, SG_LIB_CAT_ILLEGAL_REQ
- * -> bad field in cdb, -1 -> other failure */
+/* Return of 0 -> success, see sg_ll_read_capacity*() otherwise */
 int scsi_read_capacity(int sg_fd, long long * num_sect, int * sect_sz)
 {
     int k, res;
+    unsigned int ui;
     unsigned char rcBuff[RCAP16_REPLY_LEN];
-   int verb;
+    int verb;
 
     verb = (verbose ? verbose - 1: 0);
     res = sg_ll_readcap_10(sg_fd, 0, 0, rcBuff, READ_CAP_REPLY_LEN, 0,
@@ -256,8 +268,10 @@ int scsi_read_capacity(int sg_fd, long long * num_sect, int * sect_sz)
         *sect_sz = (rcBuff[8] << 24) | (rcBuff[9] << 16) |
                    (rcBuff[10] << 8) | rcBuff[11];
     } else {
-        *num_sect = 1 + ((rcBuff[0] << 24) | (rcBuff[1] << 16) |
-                    (rcBuff[2] << 8) | rcBuff[3]);
+        ui = ((rcBuff[0] << 24) | (rcBuff[1] << 16) | (rcBuff[2] << 8) |
+              rcBuff[3]);
+        /* take care not to sign extend values > 0x7fffffff */
+        *num_sect = (long long)ui + 1;
         *sect_sz = (rcBuff[4] << 24) | (rcBuff[5] << 16) |
                    (rcBuff[6] << 8) | rcBuff[7];
     }
@@ -397,26 +411,27 @@ int sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz, unsigned int blocks,
         break;
     default:
         fprintf(stderr, ME "expected cdb size of 6, 10, 12, or 16 but got"
-                        "=%d\n", cdb_sz);
+                        " %d\n", cdb_sz);
         return 1;
     }
     return 0;
 }
 
-/* -1 -> unrecoverable error, 0 -> successful, 1 -> recoverable (ENOMEM),
-   2 -> try again */
+/* 0 -> successful, SG_LIB_CAT_UNIT_ATTENTION, SG_LIB_SYNTAX_ERROR,
+ * SG_LIB_CAT_NOT_READY, SG_LIB_CAT_MEDIUM_HARD, SG_LIB_CAT_ILLEGAL_REQ,
+ * -2 -> recoverable (ENOMEM), -1 -> unrecoverable error */
 int sg_read(int sg_fd, unsigned char * buff, int blocks, long long from_block,
-            int bs, int cdbsz, int fua, int do_mmap)
+            int bs, int cdbsz, int fua, int dpo, int do_mmap)
 {
     unsigned char rdCmd[MAX_SCSI_CDBSZ];
     unsigned char senseBuff[SENSE_BUFF_LEN];
     struct sg_io_hdr io_hdr;
-    int k;
+    int k, res;
 
-    if (sg_build_scsi_cdb(rdCmd, cdbsz, blocks, from_block, 0, fua, 0)) {
+    if (sg_build_scsi_cdb(rdCmd, cdbsz, blocks, from_block, 0, fua, dpo)) {
         fprintf(stderr, ME "bad rd cdb build, from_block=%lld, blocks=%d\n",
                 from_block, blocks);
-        return -1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
     io_hdr.interface_id = 'S';
@@ -440,7 +455,9 @@ int sg_read(int sg_fd, unsigned char * buff, int blocks, long long from_block,
     }
 
 #if 1
-    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
+    while (((res = ioctl(sg_fd, SG_IO, &io_hdr)) < 0) && (EINTR == errno))
+        sleep(1);
+    if (res < 0) {
         perror(ME "SG_IO error (sg_read)");
         return -1;
     }
@@ -450,7 +467,7 @@ int sg_read(int sg_fd, unsigned char * buff, int blocks, long long from_block,
         ;
     if (res < 0) {
         if (ENOMEM == errno)
-            return 1;
+            return -2;
         perror("reading (wr) on sg device, error");
         return -1;
     }
@@ -465,17 +482,21 @@ int sg_read(int sg_fd, unsigned char * buff, int blocks, long long from_block,
 #endif
     if (verbose > 2)
         fprintf(stderr, "      duration=%u ms\n", io_hdr.duration);
-    switch (sg_err_category3(&io_hdr)) {
+    res =  sg_err_category3(&io_hdr);
+    switch (res) {
     case SG_LIB_CAT_CLEAN:
         break;
     case SG_LIB_CAT_RECOVERED:
-        sg_chk_n_print3("Reading, continuing", &io_hdr, verbose);
+        sg_chk_n_print3("Reading, continuing", &io_hdr, verbose > 1);
         break;
-    case SG_LIB_CAT_MEDIA_CHANGED:
-        return 2;
+    case SG_LIB_CAT_NOT_READY:
+    case SG_LIB_CAT_MEDIUM_HARD:
+        return res;
+    case SG_LIB_CAT_UNIT_ATTENTION:
+    case SG_LIB_CAT_ILLEGAL_REQ:
     default:
-        sg_chk_n_print3("reading", &io_hdr, verbose);
-        return -1;
+        sg_chk_n_print3("reading", &io_hdr, verbose > 1);
+        return res;
     }
     sum_of_resids += io_hdr.resid;
 #ifdef SG_DEBUG
@@ -484,20 +505,21 @@ int sg_read(int sg_fd, unsigned char * buff, int blocks, long long from_block,
     return 0;
 }
 
-/* -1 -> unrecoverable error, 0 -> successful, 1 -> recoverable (ENOMEM),
-   2 -> try again */
+/* 0 -> successful, SG_LIB_CAT_UNIT_ATTENTION, SG_LIB_SYNTAX_ERROR,
+ * SG_LIB_CAT_NOT_READY, SG_LIB_CAT_MEDIUM_HARD, SG_LIB_CAT_ILLEGAL_REQ,
+ * -2 -> recoverable (ENOMEM), -1 -> unrecoverable error */
 int sg_write(int sg_fd, unsigned char * buff, int blocks, long long to_block,
-             int bs, int cdbsz, int fua, int do_mmap, int * diop)
+             int bs, int cdbsz, int fua, int dpo, int do_mmap, int * diop)
 {
     unsigned char wrCmd[MAX_SCSI_CDBSZ];
     unsigned char senseBuff[SENSE_BUFF_LEN];
     struct sg_io_hdr io_hdr;
-    int k;
+    int k, res;
 
-    if (sg_build_scsi_cdb(wrCmd, cdbsz, blocks, to_block, 1, fua, 0)) {
+    if (sg_build_scsi_cdb(wrCmd, cdbsz, blocks, to_block, 1, fua, dpo)) {
         fprintf(stderr, ME "bad wr cdb build, to_block=%lld, blocks=%d\n",
                 to_block, blocks);
-        return -1;
+        return SG_LIB_SYNTAX_ERROR;
     }
 
     memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
@@ -524,7 +546,9 @@ int sg_write(int sg_fd, unsigned char * buff, int blocks, long long to_block,
     }
 
 #if 1
-    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
+    while (((res = ioctl(sg_fd, SG_IO, &io_hdr)) < 0) && (EINTR == errno))
+        sleep(1);
+    if (res < 0) {
         perror(ME "SG_IO error (sg_write)");
         return -1;
     }
@@ -534,7 +558,7 @@ int sg_write(int sg_fd, unsigned char * buff, int blocks, long long to_block,
         ;
     if (res < 0) {
         if (ENOMEM == errno)
-            return 1;
+            return -2;
         perror("writing (wr) on sg device, error");
         return -1;
     }
@@ -549,17 +573,21 @@ int sg_write(int sg_fd, unsigned char * buff, int blocks, long long to_block,
 #endif
     if (verbose > 2)
         fprintf(stderr, "      duration=%u ms\n", io_hdr.duration);
-    switch (sg_err_category3(&io_hdr)) {
+    res = sg_err_category3(&io_hdr);
+    switch (res) {
     case SG_LIB_CAT_CLEAN:
         break;
     case SG_LIB_CAT_RECOVERED:
-        sg_chk_n_print3("Writing, continuing", &io_hdr, verbose);
+        sg_chk_n_print3("Writing, continuing", &io_hdr, verbose > 1);
         break;
-    case SG_LIB_CAT_MEDIA_CHANGED:
-        return 2;
+    case SG_LIB_CAT_NOT_READY:
+    case SG_LIB_CAT_MEDIUM_HARD:
+        return res;
+    case SG_LIB_CAT_UNIT_ATTENTION:
+    case SG_LIB_CAT_ILLEGAL_REQ:
     default:
-        sg_chk_n_print3("writing", &io_hdr, verbose);
-        return -1;
+        sg_chk_n_print3("writing", &io_hdr, verbose > 1);
+        return res;
     }
     if (diop && *diop &&
         ((io_hdr.info & SG_INFO_DIRECT_IO_MASK) != SG_INFO_DIRECT_IO))
@@ -567,6 +595,45 @@ int sg_write(int sg_fd, unsigned char * buff, int blocks, long long to_block,
     return 0;
 }
 
+static int process_flags(const char * arg, struct flags_t * fp)
+{
+    char buff[256];
+    char * cp;
+    char * np;
+
+    strncpy(buff, arg, sizeof(buff));
+    buff[sizeof(buff) - 1] = '\0';
+    if ('\0' == buff[0]) {
+        fprintf(stderr, "no flag found\n");
+        return 1;
+    }
+    cp = buff;
+    do {
+        np = strchr(cp, ',');
+        if (np)
+            *np++ = '\0';
+        if (0 == strcmp(cp, "append"))
+            fp->append = 1;
+        else if (0 == strcmp(cp, "direct"))
+            fp->direct = 1;
+        else if (0 == strcmp(cp, "dpo"))
+            fp->dpo = 1;
+        else if (0 == strcmp(cp, "dsync"))
+            fp->dsync = 1;
+        else if (0 == strcmp(cp, "excl"))
+            fp->excl = 1;
+        else if (0 == strcmp(cp, "fua"))
+            fp->fua = 1;
+        else {
+            fprintf(stderr, "unrecognised flag: %s\n", cp);
+            return 1;
+        }
+        cp = np;
+    } while (cp);
+    return 0;
+}
+
+
 #define STR_SZ 1024
 #define INOUTF_SZ 512
 #define EBUFF_SZ 512
@@ -580,6 +647,7 @@ int main(int argc, char * argv[])
     int ibs = 0;
     int obs = 0;
     int bpt = DEF_BLOCKS_PER_TRANSFER;
+    int bpt_given = 0;
     char str[STR_SZ];
     char * key;
     char * buf;
@@ -599,109 +667,130 @@ int main(int argc, char * argv[])
     int do_time = 0;
     int scsi_cdbsz_in = DEF_SCSI_CDBSZ;
     int scsi_cdbsz_out = DEF_SCSI_CDBSZ;
+    int cdbsz_given = 0;
     int do_sync = 0;
     int do_dio = 0;
     int num_dio_not_done = 0;
-    int fua_mode = 0;
     int in_sect_sz, out_sect_sz;
+    int n, flags;
     char ebuff[EBUFF_SZ];
     int blocks_per;
     long long req_count;
     size_t psz = getpagesize();
     struct timeval start_tm, end_tm;
+    struct flags_t in_flags;
+    struct flags_t out_flags;
+    int ret = 0;
 
     inf[0] = '\0';
     outf[0] = '\0';
-    if (argc < 2) {
-        usage();
-        return 1;
-    }
+    memset(&in_flags, 0, sizeof(in_flags));
+    memset(&out_flags, 0, sizeof(out_flags));
 
-    for(k = 1; k < argc; k++) {
+    for (k = 1; k < argc; k++) {
         if (argv[k])
             strncpy(str, argv[k], STR_SZ);
         else
             continue;
-        for(key = str, buf = key; *buf && *buf != '=';)
+        for (key = str, buf = key; *buf && *buf != '=';)
             buf++;
         if (*buf)
             *buf++ = '\0';
-        if (strcmp(key,"if") == 0) {
-            if ('\0' != inf[0]) {
-                fprintf(stderr, "Second 'if=' argument??\n");
-                return 1;
-            } else
-                strncpy(inf, buf, INOUTF_SZ);
-        } else if (strcmp(key,"of") == 0) {
-            if ('\0' != outf[0]) {
-                fprintf(stderr, "Second 'of=' argument??\n");
-                return 1;
-            } else
-                strncpy(outf, buf, INOUTF_SZ);
+        if (0 == strcmp(key,"bpt")) {
+            bpt = sg_get_num(buf);
+            if (-1 == bpt) {
+                fprintf(stderr, ME "bad argument to 'bpt'\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+            bpt_given = 1;
+        } else if (0 == strcmp(key,"bs")) {
+            bs = sg_get_num(buf);
+            if (-1 == bs) {
+                fprintf(stderr, ME "bad argument to 'bs'\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+        } else if (0 == strcmp(key,"cdbsz")) {
+            scsi_cdbsz_in = sg_get_num(buf);
+            scsi_cdbsz_out = scsi_cdbsz_in;
+            cdbsz_given = 1;
+        } else if (0 == strcmp(key,"count")) {
+            dd_count = sg_get_llnum(buf);
+            if (-1LL == dd_count) {
+                fprintf(stderr, ME "bad argument to 'count'\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+        } else if (0 == strcmp(key,"dio"))
+            do_dio = sg_get_num(buf);
+        else if (0 == strcmp(key,"fua")) {
+            n = sg_get_num(buf);
+            if (n & 1)
+                out_flags.fua = 1;
+            if (n & 2)
+                in_flags.fua = 1;
         } else if (0 == strcmp(key,"ibs")) {
             ibs = sg_get_num(buf);
             if (-1 == ibs) {
                 fprintf(stderr, ME "bad argument to 'ibs'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
+            }
+        } else if (strcmp(key,"if") == 0) {
+            if ('\0' != inf[0]) {
+                fprintf(stderr, "Second 'if=' argument??\n");
+                return SG_LIB_SYNTAX_ERROR;
+            } else
+                strncpy(inf, buf, INOUTF_SZ);
+        } else if (0 == strcmp(key, "iflag")) {
+            if (process_flags(buf, &in_flags)) {
+                fprintf(stderr, ME "bad argument to 'iflag'\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+        } else if (strcmp(key,"of") == 0) {
+            if ('\0' != outf[0]) {
+                fprintf(stderr, "Second 'of=' argument??\n");
+                return SG_LIB_SYNTAX_ERROR;
+            } else
+                strncpy(outf, buf, INOUTF_SZ);
+        } else if (0 == strcmp(key, "oflag")) {
+            if (process_flags(buf, &out_flags)) {
+                fprintf(stderr, ME "bad argument to 'oflag'\n");
+                return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == strcmp(key,"obs")) {
             obs = sg_get_num(buf);
             if (-1 == obs) {
                 fprintf(stderr, ME "bad argument to 'obs'\n");
-                return 1;
-            }
-        } else if (0 == strcmp(key,"bs")) {
-            bs = sg_get_num(buf);
-            if (-1 == bs) {
-                fprintf(stderr, ME "bad argument to 'bs'\n");
-                return 1;
-            }
-        } else if (0 == strcmp(key,"bpt")) {
-            bpt = sg_get_num(buf);
-            if (-1 == bpt) {
-                fprintf(stderr, ME "bad argument to 'bpt'\n");
-                return 1;
-            }
-        } else if (0 == strcmp(key,"skip")) {
-            skip = sg_get_llnum(buf);
-            if (-1LL == skip) {
-                fprintf(stderr, ME "bad argument to 'skip'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == strcmp(key,"seek")) {
             seek = sg_get_llnum(buf);
             if (-1LL == seek) {
                 fprintf(stderr, ME "bad argument to 'seek'\n");
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
-        } else if (0 == strcmp(key,"count")) {
-            dd_count = sg_get_llnum(buf);
-            if (-1LL == dd_count) {
-                fprintf(stderr, ME "bad argument to 'count'\n");
-                return 1;
+        } else if (0 == strcmp(key,"skip")) {
+            skip = sg_get_llnum(buf);
+            if (-1LL == skip) {
+                fprintf(stderr, ME "bad argument to 'skip'\n");
+                return SG_LIB_SYNTAX_ERROR;
             }
-        } else if (0 == strcmp(key,"time"))
-            do_time = sg_get_num(buf);
-        else if (0 == strcmp(key,"cdbsz")) {
-            scsi_cdbsz_in = sg_get_num(buf);
-            scsi_cdbsz_out = scsi_cdbsz_in;
-        } else if (0 == strcmp(key,"fua"))
-            fua_mode = sg_get_num(buf);
-        else if (0 == strcmp(key,"sync"))
+        } else if (0 == strcmp(key,"sync"))
             do_sync = sg_get_num(buf);
-        else if (0 == strcmp(key,"dio"))
-            do_dio = sg_get_num(buf);
+        else if (0 == strcmp(key,"time"))
+            do_time = sg_get_num(buf);
         else if (0 == strncmp(key, "verb", 4))
             verbose = sg_get_num(buf);
-        else if (0 == strncmp(key, "--vers", 6)) {
-            fprintf(stderr, ME "for Linux sg version 3 driver: %s\n",
-                    version_str);
+        else if ((0 == strncmp(key, "--help", 7)) ||
+                 (0 == strcmp(key, "-?"))) {
+            usage();
+            return 0;
+        } else if (0 == strncmp(key, "--vers", 6)) {
+            fprintf(stderr, ME ": %s\n", version_str);
             return 0;
         }
         else {
-            fprintf(stderr, "Unrecognized argument '%s'\n", key);
-            usage();
-            return 1;
+            fprintf(stderr, "Unrecognized option '%s'\n", key);
+            fprintf(stderr, "For more information use '--help'\n");
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (bs <= 0) {
@@ -711,16 +800,26 @@ int main(int argc, char * argv[])
     if ((ibs && (ibs != bs)) || (obs && (obs != bs))) {
         fprintf(stderr, "If 'ibs' or 'obs' given must be same as 'bs'\n");
         usage();
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
     if ((skip < 0) || (seek < 0)) {
         fprintf(stderr, "skip and seek cannot be negative\n");
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
+    }
+    if ((out_flags.append > 0) && (seek > 0)) {
+        fprintf(stderr, "Can't use both append and seek switches\n");
+        return SG_LIB_SYNTAX_ERROR;
     }
     if (bpt < 1) {
         fprintf(stderr, "bpt must be greater than 0\n");
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
+    /* defaulting transfer size to 128*2048 for CD/DVDs is too large
+       for the block layer in lk 2.6 and results in an EIO on the
+       SG_IO ioctl. So reduce it in that case. */
+    if ((bs >= 2048) && (0 == bpt_given))
+        bpt = DEF_BLOCKS_PER_2048TRANSFER;
+
 #ifdef SG_DEBUG
     fprintf(stderr, ME "if=%s skip=%lld of=%s seek=%lld count=%lld\n",
            inf, skip, outf, seek, dd_count);
@@ -738,33 +837,44 @@ int main(int argc, char * argv[])
             fprintf(stderr, " >> Input file type: %s\n",
                     dd_filetype_str(in_type, ebuff));
 
-        if (FT_ST == in_type) {
+        if (FT_ERROR == in_type) {
+            fprintf(stderr, ME "unable to access %s\n", inf);
+            return SG_LIB_FILE_ERROR;
+        } else if (FT_ST == in_type) {
             fprintf(stderr, ME "unable to use scsi tape device %s\n", inf);
-            return 1;
-        }
-        else if (FT_SG == in_type) {
-            if ((infd = open(inf, O_RDWR | O_NONBLOCK)) < 0) {
+            return SG_LIB_FILE_ERROR;
+        } else if (FT_SG == in_type) {
+            flags = O_RDWR | O_NONBLOCK;
+            if (in_flags.direct)
+                flags |= O_DIRECT;
+            if (in_flags.excl)
+                flags |= O_EXCL;
+            if (in_flags.dsync)
+                flags |= O_SYNC;
+            if ((infd = open(inf, flags)) < 0) {
                 snprintf(ebuff, EBUFF_SZ, 
                          ME "could not open %s for sg reading", inf);
                 perror(ebuff);
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
             res = ioctl(infd, SG_GET_VERSION_NUM, &t);
             if ((res < 0) || (t < 30122)) {
                 fprintf(stderr, ME "sg driver prior to 3.1.22\n");
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
             in_res_sz = bs * bpt;
             if (0 != (in_res_sz % psz)) /* round up to next page */
                 in_res_sz = ((in_res_sz / psz) + 1) * psz;
             if (ioctl(infd, SG_GET_RESERVED_SIZE, &t) < 0) {
                 perror(ME "SG_GET_RESERVED_SIZE error");
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
+            if (t < MIN_RESERVED_SIZE)
+                t = MIN_RESERVED_SIZE;
             if (in_res_sz > t) {
                 if (ioctl(infd, SG_SET_RESERVED_SIZE, &in_res_sz) < 0) {
                     perror(ME "SG_SET_RESERVED_SIZE error");
-                    return 1;
+                    return SG_LIB_FILE_ERROR;
                 }
             }
             wrkMmap = mmap(NULL, in_res_sz, PROT_READ | PROT_WRITE, 
@@ -773,15 +883,22 @@ int main(int argc, char * argv[])
                 snprintf(ebuff, EBUFF_SZ,
                          ME "error using mmap() on file: %s", inf);
                 perror(ebuff);
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
         }
         else {
-            if ((infd = open(inf, O_RDONLY)) < 0) {
+            flags = O_RDONLY;
+            if (in_flags.direct)
+                flags |= O_DIRECT;
+            if (in_flags.excl)
+                flags |= O_EXCL;
+            if (in_flags.dsync)
+                flags |= O_SYNC;
+            if ((infd = open(inf, flags)) < 0) {
                 snprintf(ebuff, EBUFF_SZ,
                          ME "could not open %s for reading", inf);
                 perror(ebuff);
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
             else if (skip > 0) {
                 llse_loff_t offset = skip;
@@ -791,7 +908,7 @@ int main(int argc, char * argv[])
                     snprintf(ebuff, EBUFF_SZ, ME "couldn't skip to "
                              "required position on %s", inf);
                     perror(ebuff);
-                    return 1;
+                    return SG_LIB_FILE_ERROR;
                 }
                 if (verbose)
                     fprintf(stderr, "  >> skip: llseek SEEK_SET, "
@@ -808,29 +925,38 @@ int main(int argc, char * argv[])
 
         if (FT_ST == out_type) {
             fprintf(stderr, ME "unable to use scsi tape device %s\n", outf);
-            return 1;
+            return SG_LIB_FILE_ERROR;
         }
         else if (FT_SG == out_type) {
-            if ((outfd = open(outf, O_RDWR | O_NONBLOCK)) < 0) {
+            flags = O_RDWR | O_NONBLOCK;
+            if (out_flags.direct)
+                flags |= O_DIRECT;
+            if (out_flags.excl)
+                flags |= O_EXCL;
+            if (out_flags.dsync)
+                flags |= O_SYNC;
+            if ((outfd = open(outf, flags)) < 0) {
                 snprintf(ebuff, EBUFF_SZ, ME "could not open %s for "
                          "sg writing", outf);
                 perror(ebuff);
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
             res = ioctl(outfd, SG_GET_VERSION_NUM, &t);
             if ((res < 0) || (t < 30122)) {
                 fprintf(stderr, ME "sg driver prior to 3.1.22\n");
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
             if (ioctl(outfd, SG_GET_RESERVED_SIZE, &t) < 0) {
                 perror(ME "SG_GET_RESERVED_SIZE error");
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
+           if (t < MIN_RESERVED_SIZE)
+                t = MIN_RESERVED_SIZE;
             out_res_sz = bs * bpt;
             if (out_res_sz > t) {
                 if (ioctl(outfd, SG_SET_RESERVED_SIZE, &out_res_sz) < 0) {
                     perror(ME "SG_SET_RESERVED_SIZE error");
-                    return 1;
+                    return SG_LIB_FILE_ERROR;
                 }
             }
             if (NULL == wrkMmap) {
@@ -840,7 +966,7 @@ int main(int argc, char * argv[])
                     snprintf(ebuff, EBUFF_SZ,
                              ME "error using mmap() on file: %s", outf);
                     perror(ebuff);
-                    return 1;
+                    return SG_LIB_FILE_ERROR;
                 }
             }
         }
@@ -848,11 +974,20 @@ int main(int argc, char * argv[])
             outfd = -1; /* don't bother opening */
         else {
             if (FT_RAW != out_type) {
-                if ((outfd = open(outf, O_WRONLY | O_CREAT, 0666)) < 0) {
+                flags = O_WRONLY | O_CREAT;
+                if (out_flags.direct)
+                    flags |= O_DIRECT;
+                if (out_flags.excl)
+                    flags |= O_EXCL;
+                if (out_flags.dsync)
+                    flags |= O_SYNC;
+                if (out_flags.append)
+                    flags |= O_APPEND;
+                if ((outfd = open(outf, flags, 0666)) < 0) {
                     snprintf(ebuff, EBUFF_SZ,
                              ME "could not open %s for writing", outf);
                     perror(ebuff);
-                    return 1;
+                    return SG_LIB_FILE_ERROR;
                 }
             }
             else {
@@ -860,7 +995,7 @@ int main(int argc, char * argv[])
                     snprintf(ebuff, EBUFF_SZ, ME "could not open %s "
                              "for raw writing", outf);
                     perror(ebuff);
-                    return 1;
+                    return SG_LIB_FILE_ERROR;
                 }
             }
             if (seek > 0) {
@@ -871,7 +1006,7 @@ int main(int argc, char * argv[])
                     snprintf(ebuff, EBUFF_SZ, ME "couldn't seek to "
                              "required position on %s", outf);
                     perror(ebuff);
-                    return 1;
+                    return SG_LIB_FILE_ERROR;
                 }
                 if (verbose)
                     fprintf(stderr, "   >> seek: llseek SEEK_SET, "
@@ -882,21 +1017,25 @@ int main(int argc, char * argv[])
     if ((STDIN_FILENO == infd) && (STDOUT_FILENO == outfd)) {
         fprintf(stderr, 
                 "Can't have both 'if' as stdin _and_ 'of' as stdout\n");
-        return 1;
+        fprintf(stderr, "For more information use '--help'\n");
+        return SG_LIB_SYNTAX_ERROR;
     }
     if (dd_count < 0) {
         in_num_sect = -1;
         if (FT_SG == in_type) {
             res = scsi_read_capacity(infd, &in_num_sect, &in_sect_sz);
-            if (SG_LIB_CAT_MEDIA_CHANGED == res) {
+            if (SG_LIB_CAT_UNIT_ATTENTION == res) {
                 fprintf(stderr, 
-                        "Unit attention, media changed(in), continuing\n");
+                        "Unit attention(in), continuing\n");
                 res = scsi_read_capacity(infd, &in_num_sect, &in_sect_sz);
             }
             if (0 != res) {
-               if (res == SG_LIB_CAT_INVALID_OP)
+                if (res == SG_LIB_CAT_INVALID_OP)
                     fprintf(stderr, "read capacity not supported on %s\n",
                             inf);
+                else if (res == SG_LIB_CAT_NOT_READY)
+                    fprintf(stderr, "read capacity failed, %s not ready\n",
+                            inf);
                 else
                     fprintf(stderr, "Unable to read capacity on %s\n", inf);
                 in_num_sect = -1;
@@ -918,15 +1057,18 @@ int main(int argc, char * argv[])
         out_num_sect = -1;
         if (FT_SG == out_type) {
             res = scsi_read_capacity(outfd, &out_num_sect, &out_sect_sz);
-            if (SG_LIB_CAT_MEDIA_CHANGED == res) {
+            if (SG_LIB_CAT_UNIT_ATTENTION == res) {
                 fprintf(stderr, 
-                        "Unit attention, media changed(out), continuing\n");
+                        "Unit attention(out), continuing\n");
                 res = scsi_read_capacity(outfd, &out_num_sect, &out_sect_sz);
             }
             if (0 != res) {
-               if (res == SG_LIB_CAT_INVALID_OP)
+                if (res == SG_LIB_CAT_INVALID_OP)
                     fprintf(stderr, "read capacity not supported on %s\n",
                             outf);
+                else if (res == SG_LIB_CAT_NOT_READY)
+                    fprintf(stderr, "read capacity failed, %s not ready\n",
+                            outf);
                 else
                     fprintf(stderr, "Unable to read capacity on %s\n", outf);
                 out_num_sect = -1;
@@ -964,19 +1106,21 @@ int main(int argc, char * argv[])
 
     if (dd_count < 0) {
         fprintf(stderr, "Couldn't calculate count, please give one\n");
-        return 1;
+        return SG_LIB_SYNTAX_ERROR;
     }
-    if ((FT_SG == in_type) && ((dd_count + skip) > UINT_MAX) &&
-        (MAX_SCSI_CDBSZ != scsi_cdbsz_in)) {
-        fprintf(stderr, "Note: SCSI command size increased to 16 bytes "
-                "(for 'if')\n");
-        scsi_cdbsz_in = MAX_SCSI_CDBSZ;
-    }
-    if ((FT_SG == out_type) && ((dd_count + seek) > UINT_MAX) &&
-        (MAX_SCSI_CDBSZ != scsi_cdbsz_out)) {
-        fprintf(stderr, "Note: SCSI command size increased to 16 bytes "
-                "(for 'of')\n");
-        scsi_cdbsz_out = MAX_SCSI_CDBSZ;
+    if (! cdbsz_given) {
+        if ((FT_SG == in_type) && (MAX_SCSI_CDBSZ != scsi_cdbsz_in) &&
+            (((dd_count + skip) > UINT_MAX) || (bpt > USHRT_MAX))) {
+            fprintf(stderr, "Note: SCSI command size increased to 16 bytes "
+                    "(for 'if')\n");
+            scsi_cdbsz_in = MAX_SCSI_CDBSZ;
+        }
+        if ((FT_SG == out_type) && (MAX_SCSI_CDBSZ != scsi_cdbsz_out) &&
+            (((dd_count + seek) > UINT_MAX) || (bpt > USHRT_MAX))) {
+            fprintf(stderr, "Note: SCSI command size increased to 16 bytes "
+                    "(for 'of')\n");
+            scsi_cdbsz_out = MAX_SCSI_CDBSZ;
+        }
     }
 
     if (do_dio && (FT_SG != in_type)) {
@@ -1005,7 +1149,7 @@ int main(int argc, char * argv[])
             wrkBuff = malloc(bs * bpt + psz);
             if (0 == wrkBuff) {
                 fprintf(stderr, "Not enough user memory for raw\n");
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
             wrkPos = (unsigned char *)(((unsigned long)wrkBuff + psz - 1) &
                                        (~(psz - 1)));
@@ -1014,7 +1158,7 @@ int main(int argc, char * argv[])
             wrkBuff = malloc(bs * bpt);
             if (0 == wrkBuff) {
                 fprintf(stderr, "Not enough user memory\n");
-                return 1;
+                return SG_LIB_FILE_ERROR;
             }
             wrkPos = wrkBuff;
         }
@@ -1040,17 +1184,15 @@ int main(int argc, char * argv[])
     while (dd_count > 0) {
         blocks = (dd_count > blocks_per) ? blocks_per : dd_count;
         if (FT_SG == in_type) {
-            int fua = fua_mode & 2;
-
-            res = sg_read(infd, wrkPos, blocks, skip, bs, scsi_cdbsz_in, 
-                          fua, 1);
-            if (2 == res) {
+            ret = sg_read(infd, wrkPos, blocks, skip, bs, scsi_cdbsz_in, 
+                          in_flags.fua, in_flags.dpo, 1);
+            if (SG_LIB_CAT_UNIT_ATTENTION == ret) {
                 fprintf(stderr, 
-                        "Unit attention, media changed, continuing (r)\n");
-                res = sg_read(infd, wrkPos, blocks, skip, bs, scsi_cdbsz_in, 
-                              fua, 1);
+                        "Unit attention, continuing (r)\n");
+                ret = sg_read(infd, wrkPos, blocks, skip, bs, scsi_cdbsz_in, 
+                              in_flags.fua, in_flags.dpo, 1);
             }
-            if (0 != res) {
+            if (0 != ret) {
                 fprintf(stderr, "sg_read failed, skip=%lld\n", skip);
                 break;
             }
@@ -1064,9 +1206,10 @@ int main(int argc, char * argv[])
             if (verbose > 2)
                 fprintf(stderr, "read(unix): count=%d, res=%d\n",
                         blocks * bs, res);
-            if (res < 0) {
+            if (ret < 0) {
                 snprintf(ebuff, EBUFF_SZ, ME "reading, skip=%lld ", skip);
                 perror(ebuff);
+                ret = -1;
                 break;
             }
             else if (res < blocks * bs) {
@@ -1085,18 +1228,18 @@ int main(int argc, char * argv[])
 
         if (FT_SG == out_type) {
             int do_mmap = (FT_SG == in_type) ? 0 : 1;
-            int fua = fua_mode & 1;
             int dio_res = do_dio;
 
-            res = sg_write(outfd, wrkPos, blocks, seek, bs, scsi_cdbsz_out, 
-                           fua, do_mmap, &dio_res);
-            if (2 == res) {
+            ret = sg_write(outfd, wrkPos, blocks, seek, bs, scsi_cdbsz_out, 
+                           out_flags.fua, out_flags.dpo, do_mmap, &dio_res);
+            if (SG_LIB_CAT_UNIT_ATTENTION == ret) {
                 fprintf(stderr, 
-                        "Unit attention, media changed, continuing (w)\n");
-                res = sg_write(outfd, wrkPos, blocks, seek, bs, scsi_cdbsz_out,
-                               fua, do_mmap, &dio_res);
+                        "Unit attention, continuing (w)\n");
+                dio_res = do_dio;
+                ret = sg_write(outfd, wrkPos, blocks, seek, bs, scsi_cdbsz_out,
+                               out_flags.fua, out_flags.dpo, do_mmap, &dio_res);
             }
-            else if (0 != res) {
+            if (0 != ret) {
                 fprintf(stderr, "sg_write failed, seek=%lld\n", seek);
                 break;
             }
@@ -1161,9 +1304,8 @@ int main(int argc, char * argv[])
         if (FT_SG == out_type) {
             fprintf(stderr, ">> Synchronizing cache on %s\n", outf);
             res = sg_ll_sync_cache_10(outfd, 0, 0, 0, 0, 0, 0, 0);
-            if (2 == res) {
-                fprintf(stderr,
-                        "Unit attention, media changed(in), continuing\n");
+            if (SG_LIB_CAT_UNIT_ATTENTION == res) {
+                fprintf(stderr, "Unit attention(out), continuing\n");
                 res = sg_ll_sync_cache_10(outfd, 0, 0, 0, 0, 0, 0, 0);
             }
             if (0 != res)
@@ -1176,10 +1318,10 @@ int main(int argc, char * argv[])
         close(infd);
     if ((STDOUT_FILENO != outfd) && (FT_DEV_NULL != out_type))
         close(outfd);
-    res = 0;
     if (0 != dd_count) {
         fprintf(stderr, "Some error occurred,");
-        res = 2;
+        if (0 == ret)
+            ret = SG_LIB_CAT_OTHER;
     }
     print_stats();
     if (sum_of_resids)
@@ -1188,5 +1330,5 @@ int main(int argc, char * argv[])
     if (num_dio_not_done)
         fprintf(stderr, ">> dio requested but _not_ done %d times\n", 
                 num_dio_not_done);
-    return res;
+    return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
